"""Collection Pipeline for segmentation task."""
# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import tempfile
from copy import deepcopy
from typing import Any, Dict, List

import numpy as np
from mmcv.utils import build_from_cfg
from mmseg.datasets.builder import PIPELINES
from mmseg.datasets.pipelines import Compose
from PIL import Image
from torchvision import transforms as T
from torchvision.transforms import functional as F

from otx.algorithms.common.utils.data import get_image
from otx.api.utils.argument_checks import check_input_parameters_type

from .dataset import get_annotation_mmseg_format

_CACHE_DIR = tempfile.TemporaryDirectory(prefix="img-cache-")  # pylint: disable=consider-using-with


@PIPELINES.register_module()
class LoadImageFromOTXDataset:
    """Pipeline element that loads an image from a OTX Dataset on the fly. Can do conversion to float 32 if needed.

    Expected entries in the 'results' dict that should be passed to this pipeline element are:
        results['dataset_item']: dataset_item from which to load the image
        results['dataset_id']: id of the dataset to which the item belongs
        results['index']: index of the item in the dataset

    :param to_float32: optional bool, True to convert images to fp32. defaults to False
    """

    @check_input_parameters_type()
    def __init__(self, to_float32: bool = False):
        self.to_float32 = to_float32

    @check_input_parameters_type()
    def __call__(self, results: Dict[str, Any]):
        """Callback function LoadImageFromOTXDataset."""
        # Get image (possibly from cache)
        img = get_image(results, _CACHE_DIR.name, to_float32=self.to_float32)
        shape = img.shape

        assert img.shape[0] == results["height"], f"{img.shape[0]} != {results['height']}"
        assert img.shape[1] == results["width"], f"{img.shape[1]} != {results['width']}"

        filename = f"Dataset item index {results['index']}"
        results["filename"] = filename
        results["ori_filename"] = filename
        results["img"] = img
        results["img_shape"] = shape
        results["ori_shape"] = shape
        # Set initial values for default meta_keys
        results["pad_shape"] = shape
        num_channels = 1 if len(shape) < 3 else shape[2]
        results["img_norm_cfg"] = dict(
            mean=np.zeros(num_channels, dtype=np.float32), std=np.ones(num_channels, dtype=np.float32), to_rgb=False
        )
        results["img_fields"] = ["img"]

        return results


@PIPELINES.register_module()
class LoadAnnotationFromOTXDataset:
    """Pipeline element that loads an annotation from a OTX Dataset on the fly.

    Expected entries in the 'results' dict that should be passed to this pipeline element are:
        results['dataset_item']: dataset_item from which to load the annotation
        results['ann_info']['label_list']: list of all labels in the project

    """

    def __init__(self):
        pass

    @check_input_parameters_type()
    def __call__(self, results: Dict[str, Any]):
        """Callback function of LoadAnnotationFromOTXDataset."""
        dataset_item = results["dataset_item"]
        labels = results["ann_info"]["labels"]

        ann_info = get_annotation_mmseg_format(dataset_item, labels)

        results["gt_semantic_seg"] = ann_info["gt_semantic_seg"]
        results["seg_fields"].append("gt_semantic_seg")

        return results


@PIPELINES.register_module()
class TwoCropTransform:
    """TwoCropTransform to combine two pipelines.

    Through `TwoCropTransformHook`, how frequently both pipelines (view0 + view1) is applied can be set.

    Args:
        view0 (list): Pipeline for online network.
        view1 (list): Pipeline for target network.
    """

    def __init__(self, view0: List, view1: List):
        self.view0 = Compose([build_from_cfg(p, PIPELINES) for p in view0])
        self.view1 = Compose([build_from_cfg(p, PIPELINES) for p in view1])
        self.is_both = True

    @check_input_parameters_type()
    def __call__(self, results: Dict[str, Any]):
        """Callback function of TwoCropTransform.

        Args:
            results (dict): Inputs to be transformed.

        Returns:
            dict: Dictionary that includes stuffs for training.
                  They have different shape or attribute depending on `self.is_both`.
        """
        if self.is_both:
            results1 = self.view0(deepcopy(results))
            results2 = self.view1(deepcopy(results))

            results = deepcopy(results1)
            results["img"] = np.stack((results1["img"], results2["img"]), axis=0)
            results["gt_semantic_seg"] = np.stack((results1["gt_semantic_seg"], results2["gt_semantic_seg"]), axis=0)
            results["flip"] = [results1["flip"], results2["flip"]]

        else:
            results = self.view0(results)

        results["is_both"] = self.is_both

        return results


@PIPELINES.register_module
class RandomResizedCrop(T.RandomResizedCrop):
    """Wrapper for RandomResizedCrop in torchvision.transforms.

    Since this transformation is applied to PIL Image,
    `NDArrayToPILImage` must be applied first before this is applied.
    """

    def __call__(self, results: Dict[str, Any]):
        """Callback function of RandomResizedCrop.

        Args:
            results (dict): Inputs to be transformed.

        Returns:
            dict: Dictionary that includes transformed img and related information.
        """
        img = results["img"]
        i, j, height, width = self.get_params(img, self.scale, self.ratio)
        img = F.resized_crop(img, i, j, height, width, self.size, self.interpolation)
        results["img"] = img
        results["img_shape"] = img.size
        for key in results.get("seg_fields", []):
            results[key] = np.array(
                F.resized_crop(Image.fromarray(results[key]), i, j, height, width, self.size, self.interpolation)
            )

        # change order because of difference between numpy and PIL
        w_scale = results["img_shape"][0] / results["ori_shape"][1]
        h_scale = results["img_shape"][1] / results["ori_shape"][0]
        results["scale_factor"] = np.array([w_scale, h_scale, w_scale, h_scale], dtype=np.float32)

        return results


@PIPELINES.register_module
class RandomColorJitter(T.ColorJitter):
    """Wrapper for ColorJitter in torchvision.transforms.

    Since this transformation is applied to PIL Image,
    `NDArrayToPILImage` must be applied first before this is applied.

    Args:
        p (float): Probability for transformation. Defaults to 0.8.
    """

    def __init__(self, p: float = 0.8, **kwargs):
        super().__init__(**kwargs)
        assert 0 <= p <= 1
        self.p = p

    def __call__(self, results: Dict[str, Any]):
        """Callback function of ColorJitter.

        Args:
            results (dict): Inputs to be transformed.

        Returns:
            dict: Dictionary that includes transformed img.
        """
        if np.random.random() < self.p:
            results["img"] = self.forward(results["img"])
        return results


@PIPELINES.register_module
class RandomGrayscale(T.RandomGrayscale):
    """Wrapper for RandomGrayscale in torchvision.transforms.

    Since this transformation is applied to PIL Image,
    `NDArrayToPILImage` must be applied first before this is applied.
    """

    def __call__(self, results: Dict[str, Any]):
        """Callback function of RandomGrayscale.

        Args:
            results (dict): Inputs to be transformed.

        Returns:
            dict: Dictionary that includes transformed img.
        """
        results["img"] = self.forward(results["img"])
        return results


@PIPELINES.register_module
class RandomGaussianBlur(T.GaussianBlur):
    """Random Gaussian Blur augmentation inherited from torchvision.transforms.GaussianBlur.

    Since this transformation is applied to PIL Image,
    `NDArrayToPILImage` must be applied first before this is applied.

    Args:
        p (float): Probability for transformation. Defaults to 0.1.
    """

    def __init__(self, p: float = 0.1, **kwargs):
        super().__init__(**kwargs)
        assert 0 <= p <= 1
        self.p = p

    def __call__(self, results: Dict[str, Any]):
        """Callback function of GaussianBlur.

        Args:
            results (dict): Inputs to be transformed.

        Returns:
            dict: Dictionary that includes transformed img.
        """
        if np.random.random() < self.p:
            results["img"] = self.forward(results["img"])
        return results


@PIPELINES.register_module
class RandomSolarization:
    """Random Solarization augmentation.

    Args:
        threshold (int): Threshold for solarization. Defaults to 128.
        p (float): Probability for transformation. Defaults to 0.2.
    """

    def __init__(self, threshold: int = 128, p: float = 0.2):
        assert 0 <= p <= 1
        self.threshold = threshold
        self.p = p

    def __call__(self, results: Dict[str, Any]):
        """Callback function of Solarization.

        Args:
            results (dict): Inputs to be transformed.

        Returns:
            dict: Dictionary that includes transformed img.
        """
        if np.random.random() < self.p:
            img = results["img"]
            img = np.where(img < self.threshold, img, 255 - img)
            results["img"] = img
        return results

    def __repr__(self):
        """Set repr of Solarization."""
        repr_str = self.__class__.__name__
        return repr_str


@PIPELINES.register_module()
class NDArrayToPILImage:
    """Convert image from numpy to PIL.

    Args:
        keys (list): Keys to be transformed.
    """

    def __init__(self, keys: List[str]):
        self.keys = keys

    def __call__(self, results: Dict[str, Any]):
        """Callback function of NDArrayToPILImage.

        Args:
            results (dict): Inputs to be transformed.

        Returns:
            dict: Dictionary that includes transformed img.
        """
        for key in self.keys:
            img = results[key]
            img = Image.fromarray(img)
            results[key] = img
        return results

    def __repr__(self):
        """Set repr of NDArrayToPILImage."""
        repr_str = self.__class__.__name__
        return repr_str


@PIPELINES.register_module()
class PILImageToNDArray:
    """Convert image from PIL to numpy.

    Args:
        keys (list): Keys to be transformed.
    """

    def __init__(self, keys: List[str]):
        self.keys = keys

    def __call__(self, results: Dict[str, Any]):
        """Callback function of PILImageToNDArray.

        Args:
            results (dict): Inputs to be transformed.

        Returns:
            dict: Dictionary that includes transformed img.
        """
        for key in self.keys:
            img = results[key]
            img = np.asarray(img)
            results[key] = img
        return results

    def __repr__(self):
        """Set repr of PILImageToNDArray."""
        repr_str = self.__class__.__name__
        return repr_str



"""This module contains the interface class for tasks that can compute performance."""


# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
from typing import Optional

from otx.api.entities.resultset import ResultSetEntity


class IEvaluationTask(metaclass=abc.ABCMeta):
    """A base interface class for tasks which can compute performance on a resultset."""

    @abc.abstractmethod
    def evaluate(self, output_resultset: ResultSetEntity, evaluation_metric: Optional[str] = None):
        """Compute performance metrics for a given set of results.

        The task may use at its discretion the most appropriate metrics for the evaluation (for instance,
        average precision for classification, DICE for segmentation, etc).
        The performance will be stored directly to output_resultset.performance

        Args:
            output_resultset (ResultSetEntity): The set of results which must be
                evaluated.
            evaluation_metric (Optional[str]): the evaluation metric used to compute the
                performance
        """
        raise NotImplementedError



"""Unit Test for otx.algorithms.action.adapters.mmaction.data.cls_dataset.."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest

from otx.algorithms.action.adapters.mmaction.data.det_dataset import OTXActionDetDataset
from otx.algorithms.action.adapters.mmaction.data.pipelines import RawFrameDecode
from otx.algorithms.action.configs.detection.base.data_pipeline import train_pipeline
from otx.api.entities.label import Domain
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.action.test_helpers import (
    MockPipeline,
    generate_action_det_otx_dataset,
    generate_labels,
)


class MockDataInfoProxy(OTXActionDetDataset._DataInfoProxy):
    """Mock class for _DataInfoProxy of OTXAcitonDetDataset."""

    def __init__(self, proposals):
        self.proposals = proposals


class TestOTXActionDetDataset:
    """Test OTXActionDetDataset class.

    1. Check _DataInfoProxy
    <Steps>
        1. Create otx_dataset, labels
        2. Create _DataInfoProxy
        3. Check metadata and annotations
        4. Create proposals and check _patch_proposals
    2. Check pipelines
    3. Check loading functions
    4. Check evaluation function
    <Steps>
        1. Create sample detection inference results
        2. Check det2csv function's results
        3. Check _get_predictions function's results
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.video_len = 3
        self.frame_len = 5
        self.labels = generate_labels(3, Domain.ACTION_DETECTION)
        self.otx_dataset, self.proposals = generate_action_det_otx_dataset(self.video_len, self.frame_len, self.labels)
        self.pipeline = train_pipeline

    @e2e_pytest_unit
    def test_DataInfoProxy(self) -> None:
        """Test _DataInfoProxy in OTXActionDeteDataset."""

        proxy = OTXActionDetDataset._DataInfoProxy(self.otx_dataset, self.labels, fps=1)
        sample = proxy[0]
        assert len(proxy) == 9
        assert sample["shot_info"] == (0, 4)
        assert sample["is_empty_frame"] is False
        assert "start_index" in sample
        assert "video_id" in sample
        assert "timestamp" in sample
        assert "gt_bboxes" in sample
        assert "gt_labels" in sample

    @e2e_pytest_unit
    def test_DataInfoProxy_patch_proposals(self) -> None:
        """Test _patch_proposals function in _DataInfoProxy class."""

        proxy = MockDataInfoProxy(self.proposals)
        pre_len = len(proxy.proposals)
        proxy._patch_proposals()
        assert len(proxy.proposals) == pre_len
        for key in proxy.proposals:
            assert str(int(key.split(",")[-1])) == key.split(",")[-1]

    @e2e_pytest_unit
    def test_pipeline(self) -> None:
        """Test RawFrameDecode transform contains otx_dataset."""

        dataset = OTXActionDetDataset(self.otx_dataset, self.labels, self.pipeline, fps=1)
        for transform in dataset.pipeline.transforms:
            if isinstance(transform, RawFrameDecode):
                assert transform.otx_dataset == self.otx_dataset

    @e2e_pytest_unit
    def test_prepare_train_frames(self) -> None:
        """Test prepare_train_frames function.

        prepare_train_frames function's output should contain essential attributes for training
        """

        dataset = OTXActionDetDataset(self.otx_dataset, self.labels, self.pipeline, fps=1)
        dataset.pipeline = MockPipeline()
        sample = dataset.prepare_train_frames(0)
        assert sample["shot_info"] == (0, 4)
        assert sample["is_empty_frame"] is False
        assert "start_index" in sample
        assert "video_id" in sample
        assert "timestamp" in sample
        assert "gt_bboxes" in sample
        assert "gt_labels" in sample

    @e2e_pytest_unit
    def test_prepare_test_frames(self) -> None:
        """Test prepare_test_frames function.

        Same with test_prepare_train_frames
        """

        dataset = OTXActionDetDataset(self.otx_dataset, self.labels, self.pipeline, fps=1)
        dataset.pipeline = MockPipeline()
        sample = dataset.prepare_test_frames(0)
        assert sample["shot_info"] == (0, 4)
        assert sample["is_empty_frame"] is False
        assert "start_index" in sample
        assert "video_id" in sample
        assert "timestamp" in sample
        assert "gt_bboxes" in sample
        assert "gt_labels" in sample

    @e2e_pytest_unit
    def test_evaluate(self, mocker) -> None:
        """Test evaluate function"""

        mocker.patch(
            "otx.algorithms.action.adapters.mmaction.data.det_dataset.det_eval", return_value={"mAP@0.5IOU": 0.5}
        )
        dataset = OTXActionDetDataset(self.otx_dataset, self.labels, self.pipeline, fps=1)
        results = [
            [
                np.array([[0.0, 0.0, 1.0, 1.0, 1.0]]),
                np.array([[0.0, 0.0, 1.0, 1.0, 0.0]]),
                np.array([[0.0, 0.0, 1.0, 1.0, 0.0]]),
            ]
        ] * 9
        output = dataset.evaluate(results)
        assert output == {"mAP@0.5IOU": 0.5}



_base_ = "./segmentor.py"

__norm_cfg = dict(type="BN", requires_grad=True)
model = dict(
    type="EncoderDecoder",
    pretrained=None,
    decode_head=dict(
        type="FCNHead",
        in_channels=40,
        in_index=0,
        channels=40,
        input_transform=None,
        kernel_size=1,
        num_convs=0,
        concat_input=False,
        dropout_ratio=-1,
        num_classes=21,
        norm_cfg=__norm_cfg,
        align_corners=False,
        loss_decode=dict(type="CrossEntropyLoss"),
    ),
    # model training and testing settings
    train_cfg=dict(mix_loss=dict(enable=False, weight=0.1)),
    test_cfg=dict(mode="whole"),
)

find_unused_parameters = True



"""A file for a function get_configs_by_pairs()."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# NOTE: a workaround for https://github.com/python/mypy/issues/5028

from collections.abc import Mapping
from typing import Any, Dict, List, Literal, Sequence, Tuple, Union, overload

from mmcv import Config, ConfigDict


@overload
def get_configs_by_pairs(
    configs: Union[Config, ConfigDict, Sequence[Config], Sequence[ConfigDict]],
    pairs: Union[Dict[Any, Any], List[Dict[Any, Any]]],
    *,
    return_path: Literal[True],
) -> Dict[Tuple[Any, ...], ConfigDict]:
    ...


@overload
def get_configs_by_pairs(
    configs: Union[Config, ConfigDict, Sequence[Config], Sequence[ConfigDict]],
    pairs: Union[Dict[Any, Any], List[Dict[Any, Any]]],
    *,
    return_path: Literal[False] = False,
) -> List[ConfigDict]:
    ...


@overload
def get_configs_by_pairs(
    configs: Union[Config, ConfigDict, Sequence[Config], Sequence[ConfigDict]],
    pairs: Union[Dict[Any, Any], List[Dict[Any, Any]]],
    *,
    return_path: bool,
) -> Union[List[ConfigDict], Dict[Tuple[Any, ...], ConfigDict]]:
    ...


def get_configs_by_pairs(  # noqa: C901
    configs: Union[Config, ConfigDict, Sequence[Config], Sequence[ConfigDict]],
    pairs: Union[Dict[Any, Any], List[Dict[Any, Any]]],
    *,
    return_path: bool = False,
) -> Union[List[ConfigDict], Dict[Tuple[Any, ...], ConfigDict]]:
    """Get a list of configs by key, value pairs."""
    # TODO: multiple instance

    if not isinstance(pairs, list):
        pairs = [pairs]

    def get_config(config, path=()):
        out = {}
        if isinstance(config, (Config, Mapping)):
            for pair in pairs:
                if all(config.get(key, None) == value for key, value in pair.items()):
                    return {path: config}
            for key, value in config.items():
                out.update(get_config(value, (*path, key)))
        elif isinstance(config, (list, tuple)):
            for idx, value in enumerate(config):
                out.update(get_config(value, (*path, idx)))
        return out

    out = get_config(configs)
    if return_path:
        return out

    out_: List[ConfigDict] = []
    for found in out.values():
        if isinstance(found, (list, tuple)):
            out_.extend(found)
        else:
            out_.append(found)
    return out_



"""Anomaly Classification Task."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import ctypes
import io
import os
import shutil
import subprocess  # nosec
import tempfile
from glob import glob
from typing import Dict, List, Optional, Union

import torch
from anomalib.models import AnomalyModule, get_model
from anomalib.post_processing import NormalizationMethod, ThresholdMethod
from anomalib.utils.callbacks import (
    MetricsConfigurationCallback,
    MinMaxNormalizationCallback,
    PostProcessingConfigurationCallback,
)
from omegaconf import DictConfig, ListConfig
from pytorch_lightning import Trainer

from otx.algorithms.anomaly.adapters.anomalib.callbacks import (
    AnomalyInferenceCallback,
    ProgressCallback,
)
from otx.algorithms.anomaly.adapters.anomalib.config import get_anomalib_config
from otx.algorithms.anomaly.adapters.anomalib.data import OTXAnomalyDataModule
from otx.algorithms.anomaly.adapters.anomalib.logger import get_logger
from otx.algorithms.anomaly.configs.base.configuration import BaseAnomalyConfig
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.metrics import NullPerformance, Performance, ScoreMetric
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
    OptimizationMethod,
)
from otx.api.entities.model_template import TaskType
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.evaluation.performance_provider_interface import (
    IPerformanceProvider,
)
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.export_interface import ExportType, IExportTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.unload_interface import IUnload

logger = get_logger(__name__)


# pylint: disable=too-many-instance-attributes
class InferenceTask(IInferenceTask, IEvaluationTask, IExportTask, IUnload):
    """Base Anomaly Task."""

    def __init__(self, task_environment: TaskEnvironment, output_path: Optional[str] = None) -> None:
        """Train, Infer, Export, Optimize and Deploy an Anomaly Classification Task.

        Args:
            task_environment (TaskEnvironment): OTX Task environment.
            output_path (Optional[str]): output path where task output are saved.
        """
        torch.backends.cudnn.enabled = True
        logger.info("Initializing the task environment.")
        self.task_environment = task_environment
        self.task_type = task_environment.model_template.task_type
        self.model_name = task_environment.model_template.name
        self.labels = task_environment.get_labels()

        template_file_path = task_environment.model_template.model_template_path
        self.base_dir = os.path.abspath(os.path.dirname(template_file_path))

        # Hyperparameters.
        self._work_dir_is_temp = False
        if output_path is None:
            output_path = tempfile.mkdtemp(prefix="otx-anomalib")
            self._work_dir_is_temp = True
        self.project_path: str = output_path
        self.config = self.get_config()

        # Set default model attributes.
        self.optimization_methods: List[OptimizationMethod] = []
        self.precision = [ModelPrecision.FP32]
        self.optimization_type = ModelOptimizationType.MO

        self.model = self.load_model(otx_model=task_environment.model)

        self.trainer: Trainer

    def get_config(self) -> Union[DictConfig, ListConfig]:
        """Get Anomalib Config from task environment.

        Returns:
            Union[DictConfig, ListConfig]: Anomalib config.
        """
        self.hyper_parameters: BaseAnomalyConfig = self.task_environment.get_hyper_parameters()
        config = get_anomalib_config(task_name=self.model_name, otx_config=self.hyper_parameters)
        config.project.path = self.project_path

        config.dataset.task = "classification"

        return config

    def load_model(self, otx_model: Optional[ModelEntity]) -> AnomalyModule:
        """Create and Load Anomalib Module from OTX Model.

        This method checks if the task environment has a saved OTX Model,
        and creates one. If the OTX model already exists, it returns the
        the model with the saved weights.

        Args:
            otx_model (Optional[ModelEntity]): OTX Model from the
                task environment.

        Returns:
            AnomalyModule: Anomalib
                classification or segmentation model with/without weights.
        """
        if otx_model is None:
            model = get_model(config=self.config)
            logger.info(
                "No trained model in project yet. Created new model with '%s'",
                self.model_name,
            )
        else:
            buffer = io.BytesIO(otx_model.get_data("weights.pth"))
            model_data = torch.load(buffer, map_location=torch.device("cpu"))

            if model_data["config"]["model"]["backbone"] != self.config["model"]["backbone"]:
                logger.warning(
                    "Backbone of the model in the Task Environment is different from the one in the template. "
                    f"creating model with backbone={model_data['config']['model']['backbone']}"
                )
                self.config["model"]["backbone"] = model_data["config"]["model"]["backbone"]
            try:
                model = get_model(config=self.config)
                model.load_state_dict(model_data["model"])
                logger.info("Loaded model weights from Task Environment")
            except BaseException as exception:
                raise ValueError("Could not load the saved model. The model file structure is invalid.") from exception

        return model

    def cancel_training(self) -> None:
        """Cancel the training `after_batch_end`.

        This terminates the training; however validation is still performed.
        """
        logger.info("Cancel training requested.")
        self.trainer.should_stop = True

        # The runner periodically checks `.stop_training` file to ensure if cancellation is requested.
        cancel_training_file_path = os.path.join(self.config.project.path, ".stop_training")
        with open(file=cancel_training_file_path, mode="a", encoding="utf-8"):
            pass

    def infer(self, dataset: DatasetEntity, inference_parameters: InferenceParameters) -> DatasetEntity:
        """Perform inference on a dataset.

        Args:
            dataset (DatasetEntity): Dataset to infer.
            inference_parameters (InferenceParameters): Inference parameters.

        Returns:
            DatasetEntity: Output dataset with predictions.
        """
        logger.info("Performing inference on the validation set using the base torch model.")
        config = self.get_config()
        datamodule = OTXAnomalyDataModule(config=config, dataset=dataset, task_type=self.task_type)

        logger.info("Inference Configs '%s'", config)

        # Callbacks.
        progress = ProgressCallback(parameters=inference_parameters)
        inference = AnomalyInferenceCallback(dataset, self.labels, self.task_type)
        normalize = MinMaxNormalizationCallback()
        metrics_configuration = MetricsConfigurationCallback(
            task=config.dataset.task,
            image_metrics=config.metrics.image,
            pixel_metrics=config.metrics.get("pixel"),
        )
        post_processing_configuration = PostProcessingConfigurationCallback(
            normalization_method=NormalizationMethod.MIN_MAX,
            threshold_method=ThresholdMethod.ADAPTIVE,
            manual_image_threshold=config.metrics.threshold.manual_image,
            manual_pixel_threshold=config.metrics.threshold.manual_pixel,
        )
        callbacks = [progress, normalize, inference, metrics_configuration, post_processing_configuration]

        self.trainer = Trainer(**config.trainer, logger=False, callbacks=callbacks)
        self.trainer.predict(model=self.model, datamodule=datamodule)
        return dataset

    def evaluate(self, output_resultset: ResultSetEntity, evaluation_metric: Optional[str] = None) -> None:
        """Evaluate the performance on a result set.

        Args:
            output_resultset (ResultSetEntity): Result Set from which the performance is evaluated.
            evaluation_metric (Optional[str], optional): Evaluation metric. Defaults to None. Instead,
                metric is chosen depending on the task type.
        """
        metric: IPerformanceProvider
        if self.task_type == TaskType.ANOMALY_CLASSIFICATION:
            metric = MetricsHelper.compute_f_measure(output_resultset)
        elif self.task_type == TaskType.ANOMALY_DETECTION:
            metric = MetricsHelper.compute_anomaly_detection_scores(output_resultset)
        elif self.task_type == TaskType.ANOMALY_SEGMENTATION:
            metric = MetricsHelper.compute_anomaly_segmentation_scores(output_resultset)
        else:
            raise ValueError(f"Unknown task type: {self.task_type}")
        output_resultset.performance = metric.get_performance()

        if self.task_type == TaskType.ANOMALY_CLASSIFICATION:
            accuracy = MetricsHelper.compute_accuracy(output_resultset).get_performance()
            output_resultset.performance.dashboard_metrics.extend(accuracy.dashboard_metrics)

    def _export_to_onnx(self, onnx_path: str):
        """Export model to ONNX.

        Args:
             onnx_path (str): path to save ONNX file
        """
        height, width = self.config.model.input_size
        torch.onnx.export(
            model=self.model.model,
            args=torch.zeros((1, 3, height, width)).to(self.model.device),
            f=onnx_path,
            opset_version=11,
        )

    def export(self, export_type: ExportType, output_model: ModelEntity, dump_features: bool = True) -> None:
        """Export model to OpenVINO IR.

        Args:
            export_type (ExportType): Export type should be ExportType.OPENVINO
            output_model (ModelEntity): The model entity in which to write the OpenVINO IR data
            dump_features (bool): Flag to return "feature_vector" and "saliency_map".

        Raises:
            Exception: If export_type is not ExportType.OPENVINO
        """
        # TODO: add dumping saliency maps and representation vectors according to dump_features flag
        if not dump_features:
            logger.warning(
                "Ommitting feature dumping is not implemented."
                "The saliency maps and representation vector outputs will be dumped in the exported model."
            )

        assert export_type == ExportType.OPENVINO, f"Incorrect export_type={export_type}"
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.MO

        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = self.optimization_type

        # pylint: disable=no-member; need to refactor this
        logger.info("Exporting the OpenVINO model.")
        onnx_path = os.path.join(self.config.project.path, "onnx_model.onnx")
        self._export_to_onnx(onnx_path)
        optimize_command = ["mo", "--input_model", onnx_path, "--output_dir", self.config.project.path]
        subprocess.run(optimize_command, check=True)
        bin_file = glob(os.path.join(self.config.project.path, "*.bin"))[0]
        xml_file = glob(os.path.join(self.config.project.path, "*.xml"))[0]
        with open(bin_file, "rb") as file:
            output_model.set_data("openvino.bin", file.read())
        with open(xml_file, "rb") as file:
            output_model.set_data("openvino.xml", file.read())

        output_model.precision = self.precision
        output_model.optimization_methods = self.optimization_methods

        output_model.set_data("label_schema.json", label_schema_to_bytes(self.task_environment.label_schema))
        self._set_metadata(output_model)

    def model_info(self) -> Dict:
        """Return model info to save the model weights.

        Returns:
           Dict: Model info.
        """
        return {
            "model": self.model.state_dict(),
            "config": self.get_config(),
            "VERSION": 1,
        }

    def save_model(self, output_model: ModelEntity) -> None:
        """Save the model after training is completed.

        Args:
            output_model (ModelEntity): Output model onto which the weights are saved.
        """
        logger.info("Saving the model weights.")
        model_info = self.model_info()
        buffer = io.BytesIO()
        torch.save(model_info, buffer)
        output_model.set_data("weights.pth", buffer.getvalue())
        output_model.set_data("label_schema.json", label_schema_to_bytes(self.task_environment.label_schema))
        self._set_metadata(output_model)

        if hasattr(self.model, "image_metrics"):
            f1_score = self.model.image_metrics.F1Score.compute().item()
            output_model.performance = Performance(score=ScoreMetric(name="F1 Score", value=f1_score))
        else:
            output_model.performance = NullPerformance()
        output_model.precision = self.precision
        output_model.optimization_methods = self.optimization_methods

    def _set_metadata(self, output_model: ModelEntity):
        if hasattr(self.model, "image_threshold"):
            output_model.set_data("image_threshold", self.model.image_threshold.value.cpu().numpy().tobytes())
        if hasattr(self.model, "pixel_threshold"):
            output_model.set_data("pixel_threshold", self.model.pixel_threshold.value.cpu().numpy().tobytes())
        if hasattr(self.model, "normalization_metrics"):
            output_model.set_data("min", self.model.normalization_metrics.state_dict()["min"].cpu().numpy().tobytes())
            output_model.set_data("max", self.model.normalization_metrics.state_dict()["max"].cpu().numpy().tobytes())
        else:
            logger.warning(
                "The model was not trained before saving. This will lead to incorrect normalization of the heatmaps."
            )

    @staticmethod
    def _is_docker() -> bool:
        """Check whether the task runs in docker container.

        Returns:
            bool: True if task runs in docker, False otherwise.
        """
        path = "/proc/self/cgroup"
        is_in_docker = False
        if os.path.isfile(path):
            with open(path, encoding="utf-8") as file:
                is_in_docker = is_in_docker or any("docker" in line for line in file)
        is_in_docker = is_in_docker or os.path.exists("/.dockerenv")
        return is_in_docker

    def unload(self) -> None:
        """Unload the task."""
        self.cleanup()

        if self._is_docker():
            logger.warning("Got unload request. Unloading models. Throwing Segmentation Fault on purpose")
            ctypes.string_at(0)

        else:
            logger.warning("Got unload request, but not on Docker. Only clearing CUDA cache")
            torch.cuda.empty_cache()
            logger.warning(
                "Done unloading. Torch is still occupying %f bytes of GPU memory",
                torch.cuda.memory_allocated(),
            )

    def cleanup(self) -> None:
        """Clean up work directory."""
        if self._work_dir_is_temp and os.path.exists(self.config.project.path):
            shutil.rmtree(self.config.project.path, ignore_errors=False)



"""Anomaly Classification Task."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
import json
import os
import re
from typing import Dict, Optional

import torch
from anomalib.models import AnomalyModule, get_model
from anomalib.post_processing import NormalizationMethod, ThresholdMethod
from anomalib.utils.callbacks import (
    MetricsConfigurationCallback,
    MinMaxNormalizationCallback,
    PostProcessingConfigurationCallback,
)
from anomalib.utils.callbacks.nncf.callback import NNCFCallback
from anomalib.utils.callbacks.nncf.utils import (
    compose_nncf_config,
    is_state_nncf,
    wrap_nncf_model,
)
from pytorch_lightning import Trainer

from otx.algorithms.anomaly.adapters.anomalib.callbacks import ProgressCallback
from otx.algorithms.anomaly.adapters.anomalib.data import OTXAnomalyDataModule
from otx.algorithms.anomaly.adapters.anomalib.logger import get_logger
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
    OptimizationMethod,
)
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.usecases.tasks.interfaces.optimization_interface import (
    IOptimizationTask,
    OptimizationType,
)

from .inference import InferenceTask

logger = get_logger(__name__)


class NNCFTask(InferenceTask, IOptimizationTask):
    """Base Anomaly Task."""

    def __init__(self, task_environment: TaskEnvironment, **kwargs) -> None:
        """Task for compressing models using NNCF.

        Args:
            task_environment (TaskEnvironment): OTX Task environment.
        """
        self.compression_ctrl = None
        self.nncf_preset = "nncf_quantization"
        super().__init__(task_environment, **kwargs)
        self.optimization_type = ModelOptimizationType.NNCF

    def _set_attributes_by_hyperparams(self):
        quantization = self.hyper_parameters.nncf_optimization.enable_quantization
        pruning = self.hyper_parameters.nncf_optimization.enable_pruning
        if quantization and pruning:
            self.nncf_preset = "nncf_quantization_pruning"
            self.optimization_methods = [
                OptimizationMethod.QUANTIZATION,
                OptimizationMethod.FILTER_PRUNING,
            ]
            self.precision = [ModelPrecision.INT8]
            return
        if quantization and not pruning:
            self.nncf_preset = "nncf_quantization"
            self.optimization_methods = [OptimizationMethod.QUANTIZATION]
            self.precision = [ModelPrecision.INT8]
            return
        if not quantization and pruning:
            self.nncf_preset = "nncf_pruning"
            self.optimization_methods = [OptimizationMethod.FILTER_PRUNING]
            self.precision = [ModelPrecision.FP32]
            return
        raise RuntimeError("Not selected optimization algorithm")

    def load_model(self, otx_model: Optional[ModelEntity]) -> AnomalyModule:
        """Create and Load Anomalib Module from OTX Model.

        This method checks if the task environment has a saved OTX Model,
        and creates one. If the OTX model already exists, it returns the
        the model with the saved weights.

        Args:
            otx_model (Optional[ModelEntity]): OTX Model from the
                task environment.

        Returns:
            AnomalyModule: Anomalib
                classification or segmentation model with/without weights.
        """
        nncf_config_path = os.path.join(self.base_dir, "compression_config.json")

        with open(nncf_config_path, encoding="utf8") as nncf_config_file:
            common_nncf_config = json.load(nncf_config_file)

        self._set_attributes_by_hyperparams()
        self.optimization_config = compose_nncf_config(common_nncf_config, [self.nncf_preset])
        self.config.merge_with(self.optimization_config)
        model = get_model(config=self.config)
        if otx_model is None:
            raise ValueError("No trained model in project. NNCF require pretrained weights to compress the model")

        buffer = io.BytesIO(otx_model.get_data("weights.pth"))  # type: ignore
        model_data = torch.load(buffer, map_location=torch.device("cpu"))

        if is_state_nncf(model_data):
            logger.info("Loaded model weights from Task Environment and wrapped by NNCF")

            # Fix name mismatch for wrapped model by pytorch_lighting
            nncf_modules = {}
            pl_modules = {}
            for key in model_data["model"].keys():
                if key.startswith("model."):
                    new_key = key.replace("model.", "")
                    res = re.search(r"nncf_module\.(\w+)_feature_extractor\.(.*)", new_key)
                    if res:
                        new_key = f"nncf_module.{res.group(1)}_model.feature_extractor.{res.group(2)}"
                    nncf_modules[new_key] = model_data["model"][key]
                else:
                    pl_modules[key] = model_data["model"][key]
            model_data["model"] = nncf_modules

            self.compression_ctrl, model.model = wrap_nncf_model(
                model.model,
                self.optimization_config["nncf_config"],
                init_state_dict=model_data,
            )
            # Load extra parameters of pytorch_lighting model
            model.load_state_dict(pl_modules, strict=False)
        else:
            try:
                model.load_state_dict(model_data["model"])
                logger.info("Loaded model weights from Task Environment")
            except BaseException as exception:
                raise ValueError("Could not load the saved model. The model file structure is invalid.") from exception

        return model

    def optimize(
        self,
        optimization_type: OptimizationType,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        optimization_parameters: Optional[OptimizationParameters] = None,
    ):
        """Train the anomaly classification model.

        Args:
            optimization_type (OptimizationType): Type of optimization.
            dataset (DatasetEntity): Input dataset.
            output_model (ModelEntity): Output model to save the model weights.
            optimization_parameters (OptimizationParameters): Training parameters
        """
        logger.info("Optimization the model.")

        if optimization_type is not OptimizationType.NNCF:
            raise RuntimeError("NNCF is the only supported optimization")

        datamodule = OTXAnomalyDataModule(config=self.config, dataset=dataset, task_type=self.task_type)
        nncf_callback = NNCFCallback(config=self.optimization_config["nncf_config"])
        metrics_configuration = MetricsConfigurationCallback(
            task=self.config.dataset.task,
            image_metrics=self.config.metrics.image,
            pixel_metrics=self.config.metrics.get("pixel"),
        )
        post_processing_configuration = PostProcessingConfigurationCallback(
            normalization_method=NormalizationMethod.MIN_MAX,
            threshold_method=ThresholdMethod.ADAPTIVE,
            manual_image_threshold=self.config.metrics.threshold.manual_image,
            manual_pixel_threshold=self.config.metrics.threshold.manual_pixel,
        )
        callbacks = [
            ProgressCallback(parameters=optimization_parameters),
            MinMaxNormalizationCallback(),
            nncf_callback,
            metrics_configuration,
            post_processing_configuration,
        ]

        self.trainer = Trainer(**self.config.trainer, logger=False, callbacks=callbacks)
        self.trainer.fit(model=self.model, datamodule=datamodule)
        self.compression_ctrl = nncf_callback.nncf_ctrl
        output_model.model_format = ModelFormat.BASE_FRAMEWORK
        output_model.optimization_type = ModelOptimizationType.NNCF
        self.save_model(output_model)

        logger.info("Training completed.")

    def model_info(self) -> Dict:
        """Return model info to save the model weights.

        Returns:
           Dict: Model info.
        """
        return {
            "compression_state": self.compression_ctrl.get_compression_state(),  # type: ignore
            "meta": {
                "config": self.config,
                "nncf_enable_compression": True,
            },
            "model": self.model.state_dict(),
            "config": self.get_config(),
            "VERSION": 1,
        }

    def _export_to_onnx(self, onnx_path: str):
        """Export model to ONNX.

        Args:
             onnx_path (str): path to save ONNX file
        """
        self.compression_ctrl.export_model(onnx_path, "onnx_11")  # type: ignore



"""This module implements the Graph interface."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import abc
from typing import Iterator, List

import networkx as nx


class IGraph(metaclass=abc.ABCMeta):
    """This interface describes how the interface of the Graph looks like.

    This interface is used to represent the TaskGraph inside project as well as the label tree inside LabelSchema
    """

    @abc.abstractmethod
    def add_node(self, node):
        """Add node to the graph."""
        raise NotImplementedError

    @abc.abstractmethod
    def add_edge(self, node1, node2):
        """Add an edge between node1 and node2."""
        raise NotImplementedError

    @abc.abstractmethod
    def has_edge_between(self, node1, node2):
        """Returns whether there is an edge between `node1` and `node2`."""
        raise NotImplementedError

    @abc.abstractmethod
    def neighbors(self, node) -> List[dict]:
        """Returns neighbors of `node`."""
        raise NotImplementedError

    @abc.abstractmethod
    def find_cliques(self) -> Iterator[List[dict]]:
        """Returns cliques in the graph."""
        raise NotImplementedError

    @abc.abstractmethod
    def num_nodes(self) -> int:
        """Returns number of nodes."""
        raise NotImplementedError

    @abc.abstractmethod
    def remove_edges(self, node1, node2) -> None:
        """Removes the edges between two nodes."""
        raise NotImplementedError

    @abc.abstractmethod
    def find_out_edges(self, node) -> nx.reportviews.OutMultiEdgeView:
        """Returns the edges coming out of the node."""
        raise NotImplementedError

    @abc.abstractmethod
    def find_in_edges(self, node) -> nx.reportviews.InMultiEdgeView:
        """Returns the edges coming in to the node."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def edges(self) -> nx.reportviews.OutMultiEdgeView:
        """Returns the edges in the Graph."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def nodes(self) -> nx.reportviews.NodeView:
        """Return nodes in the graph."""
        raise NotImplementedError



"""OpenVINO Anomaly Task."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
import json
import os
import tempfile
from typing import Any, Dict, List, Optional
from zipfile import ZipFile

import numpy as np
from addict import Dict as ADDict
from anomalib.deploy import OpenVINOInferencer
from anomalib.post_processing import anomaly_map_to_color_map
from compression.api import DataLoader
from compression.engines.ie_engine import IEEngine
from compression.graph import load_model, save_model
from compression.graph.model_utils import compress_model_weights, get_nodes_by_type
from compression.pipeline.initializer import create_pipeline
from omegaconf import OmegaConf

import otx.algorithms.anomaly.adapters.anomalib.exportable_code
from otx.algorithms.anomaly.adapters.anomalib.config import get_anomalib_config
from otx.algorithms.anomaly.adapters.anomalib.logger import get_logger
from otx.algorithms.anomaly.configs.base.configuration import BaseAnomalyConfig
from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import (
    InferenceParameters,
    default_progress_callback,
)
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
    OptimizationMethod,
)
from otx.api.entities.model_template import TaskType
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.result_media import ResultMediaEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.serialization.label_mapper import LabelSchemaMapper, label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.evaluation.performance_provider_interface import (
    IPerformanceProvider,
)
from otx.api.usecases.exportable_code import demo
from otx.api.usecases.tasks.interfaces.deployment_interface import IDeploymentTask
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.optimization_interface import (
    IOptimizationTask,
    OptimizationType,
)
from otx.api.utils.anomaly_utils import create_detection_annotation_from_anomaly_heatmap
from otx.api.utils.segmentation_utils import create_annotation_from_segmentation_map

logger = get_logger(__name__)


class OTXOpenVINOAnomalyDataloader(DataLoader):
    """Dataloader for loading OTX dataset into OTX OpenVINO Inferencer.

    Args:
        dataset (DatasetEntity): OTX dataset entity
        inferencer (OpenVINOInferencer): OpenVINO Inferencer
    """

    def __init__(
        self,
        config: ADDict,
        dataset: DatasetEntity,
        inferencer: OpenVINOInferencer,
    ):
        super().__init__(config=config)
        self.dataset = dataset
        self.inferencer = inferencer

    def __getitem__(self, index: int):
        """Get dataset item.

        Args:
            index (int): Index of the dataset sample.

        Returns:
            Dataset item.
        """
        image = self.dataset[index].numpy
        annotation = self.dataset[index].annotation_scene
        inputs = self.inferencer.pre_process(image)

        return (index, annotation), inputs

    def __len__(self) -> int:
        """Get size of the dataset.

        Returns:
            int: Size of the dataset.
        """
        return len(self.dataset)


class OpenVINOTask(IInferenceTask, IEvaluationTask, IOptimizationTask, IDeploymentTask):
    """OpenVINO inference task.

    Args:
        task_environment (TaskEnvironment): task environment of the trained anomaly model
    """

    def __init__(self, task_environment: TaskEnvironment) -> None:
        logger.info("Initializing the OpenVINO task.")
        self.task_environment = task_environment
        self.task_type = self.task_environment.model_template.task_type
        self.config = self.get_config()
        self.inferencer = self.load_inferencer()

        labels = self.task_environment.get_labels()
        self.normal_label = [label for label in labels if not label.is_anomalous][0]
        self.anomalous_label = [label for label in labels if label.is_anomalous][0]

        template_file_path = task_environment.model_template.model_template_path
        self._base_dir = os.path.abspath(os.path.dirname(template_file_path))

    def get_config(self) -> ADDict:
        """Get Anomalib Config from task environment.

        Returns:
            ADDict: Anomalib config
        """
        task_name = self.task_environment.model_template.name
        otx_config: ConfigurableParameters = self.task_environment.get_hyper_parameters()
        config = get_anomalib_config(task_name=task_name, otx_config=otx_config)
        return ADDict(OmegaConf.to_container(config))

    def infer(self, dataset: DatasetEntity, inference_parameters: InferenceParameters) -> DatasetEntity:
        """Perform Inference.

        Args:
            dataset (DatasetEntity): Inference dataset
            inference_parameters (InferenceParameters): Inference parameters.

        Returns:
            DatasetEntity: Output dataset storing inference predictions.
        """
        if self.task_environment.model is None:
            raise Exception("task_environment.model is None. Cannot access threshold to calculate labels.")

        logger.info("Start OpenVINO inference.")
        update_progress_callback = default_progress_callback
        if inference_parameters is not None:
            update_progress_callback = inference_parameters.update_progress  # type: ignore

        # This always assumes that threshold is available in the task environment's model
        meta_data = self.get_meta_data()
        for idx, dataset_item in enumerate(dataset):
            image_result = self.inferencer.predict(dataset_item.numpy, meta_data=meta_data)

            # TODO: inferencer should return predicted label and mask
            pred_label = image_result.pred_score >= 0.5
            pred_mask = (image_result.anomaly_map >= 0.5).astype(np.uint8)
            probability = image_result.pred_score if pred_label else 1 - image_result.pred_score
            if self.task_type == TaskType.ANOMALY_CLASSIFICATION:
                label = self.anomalous_label if image_result.pred_score >= 0.5 else self.normal_label
            elif self.task_type == TaskType.ANOMALY_SEGMENTATION:
                annotations = create_annotation_from_segmentation_map(
                    pred_mask, image_result.anomaly_map.squeeze(), {0: self.normal_label, 1: self.anomalous_label}
                )
                dataset_item.append_annotations(annotations)
                label = self.normal_label if len(annotations) == 0 else self.anomalous_label
            elif self.task_type == TaskType.ANOMALY_DETECTION:
                annotations = create_detection_annotation_from_anomaly_heatmap(
                    pred_mask, image_result.anomaly_map.squeeze(), {0: self.normal_label, 1: self.anomalous_label}
                )
                dataset_item.append_annotations(annotations)
                label = self.normal_label if len(annotations) == 0 else self.anomalous_label
            else:
                raise ValueError(f"Unknown task type: {self.task_type}")

            dataset_item.append_labels([ScoredLabel(label=label, probability=float(probability))])
            anomaly_map = anomaly_map_to_color_map(image_result.anomaly_map, normalize=False)
            heatmap_media = ResultMediaEntity(
                name="Anomaly Map",
                type="anomaly_map",
                annotation_scene=dataset_item.annotation_scene,
                numpy=anomaly_map,
            )
            dataset_item.append_metadata_item(heatmap_media)
            update_progress_callback(int((idx + 1) / len(dataset) * 100))

        return dataset

    def get_meta_data(self) -> Dict:
        """Get Meta Data."""
        meta_data = {}
        if self.task_environment.model is not None:
            model: ModelEntity = self.task_environment.model
            image_threshold = np.frombuffer(model.get_data("image_threshold"), dtype=np.float32)
            pixel_threshold = np.frombuffer(model.get_data("pixel_threshold"), dtype=np.float32)
            min_value = np.frombuffer(model.get_data("min"), dtype=np.float32)
            max_value = np.frombuffer(model.get_data("max"), dtype=np.float32)
            meta_data = {
                "image_threshold": image_threshold,
                "pixel_threshold": pixel_threshold,
                "min": min_value,
                "max": max_value,
            }
        else:
            raise ValueError("Cannot access meta-data. self.task_environment.model is empty.")

        return meta_data

    def evaluate(self, output_resultset: ResultSetEntity, evaluation_metric: Optional[str] = None):
        """Evaluate the performance of the model.

        Args:
            output_resultset (ResultSetEntity): Result set storing ground truth and predicted dataset.
            evaluation_metric (Optional[str], optional): Evaluation metric. Defaults to None.
        """
        metric: IPerformanceProvider
        if self.task_type == TaskType.ANOMALY_CLASSIFICATION:
            metric = MetricsHelper.compute_f_measure(output_resultset)
        elif self.task_type == TaskType.ANOMALY_DETECTION:
            metric = MetricsHelper.compute_anomaly_detection_scores(output_resultset)
        elif self.task_type == TaskType.ANOMALY_SEGMENTATION:
            metric = MetricsHelper.compute_anomaly_segmentation_scores(output_resultset)
        else:
            raise ValueError(f"Unknown task type: {self.task_type}")
        output_resultset.performance = metric.get_performance()

    def _get_optimization_algorithms_configs(self) -> List[ADDict]:
        """Returns list of optimization algorithms configurations."""
        hparams: BaseAnomalyConfig = self.task_environment.get_hyper_parameters()

        optimization_config_path = os.path.join(self._base_dir, "pot_optimization_config.json")
        if os.path.exists(optimization_config_path):
            with open(optimization_config_path, encoding="UTF-8") as f_src:
                algorithms = ADDict(json.load(f_src))["algorithms"]
        else:
            algorithms = [
                ADDict({"name": "DefaultQuantization", "params": {"target_device": "ANY", "shuffle_data": True}})
            ]
        for algo in algorithms:
            algo.params.stat_subset_size = hparams.pot_parameters.stat_subset_size
            algo.params.shuffle_data = True
            if "Quantization" in algo["name"]:
                algo.params.preset = hparams.pot_parameters.preset.name.lower()

        return algorithms

    def optimize(
        self,
        optimization_type: OptimizationType,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        optimization_parameters: Optional[OptimizationParameters],
    ):
        """Optimize the model.

        Args:
            optimization_type (OptimizationType): Type of optimization [POT or NNCF]
            dataset (DatasetEntity): Input Dataset.
            output_model (ModelEntity): Output model.
            optimization_parameters (Optional[OptimizationParameters]): Optimization parameters.

        Raises:
            ValueError: When the optimization type is not POT, which is the only support type at the moment.
        """
        if optimization_type is not OptimizationType.POT:
            raise ValueError("POT is the only supported optimization type for OpenVINO models")

        # Training subset does not contain example of anomalous images.
        # Anomalous examples from all dataset used to get statistics for quantization.
        dataset = DatasetEntity(
            items=[item for item in dataset if item.get_shapes_labels()[0].is_anomalous], purpose=dataset.purpose
        )

        logger.info("Starting POT optimization.")
        data_loader = OTXOpenVINOAnomalyDataloader(config=self.config, dataset=dataset, inferencer=self.inferencer)

        with tempfile.TemporaryDirectory() as tempdir:
            xml_path = os.path.join(tempdir, "model.xml")
            bin_path = os.path.join(tempdir, "model.bin")

            if self.task_environment.model is not None:
                self.__save_weights(xml_path, self.task_environment.model.get_data("openvino.xml"))
                self.__save_weights(bin_path, self.task_environment.model.get_data("openvino.bin"))
            else:
                raise ValueError("Cannot save the weights. self.task_environment.model is None.")

            model_config = {
                "model_name": "openvino_model",
                "model": xml_path,
                "weights": bin_path,
            }
            model = load_model(model_config)

            if get_nodes_by_type(model, ["FakeQuantize"]):
                raise RuntimeError("Model is already optimized by POT")

        if optimization_parameters is not None:
            optimization_parameters.update_progress(10, None)

        engine = IEEngine(config=ADDict({"device": "CPU"}), data_loader=data_loader, metric=None)
        pipeline = create_pipeline(algo_config=self._get_optimization_algorithms_configs(), engine=engine)
        compressed_model = pipeline.run(model)
        compress_model_weights(compressed_model)

        if optimization_parameters is not None:
            optimization_parameters.update_progress(90, None)

        with tempfile.TemporaryDirectory() as tempdir:
            save_model(compressed_model, tempdir, model_name="model")
            self.__load_weights(path=os.path.join(tempdir, "model.xml"), output_model=output_model, key="openvino.xml")
            self.__load_weights(path=os.path.join(tempdir, "model.bin"), output_model=output_model, key="openvino.bin")

        output_model.set_data("label_schema.json", label_schema_to_bytes(self.task_environment.label_schema))
        output_model.set_data("image_threshold", self.task_environment.model.get_data("image_threshold"))
        output_model.set_data("pixel_threshold", self.task_environment.model.get_data("pixel_threshold"))
        output_model.set_data("min", self.task_environment.model.get_data("min"))
        output_model.set_data("max", self.task_environment.model.get_data("max"))
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.POT
        output_model.optimization_methods = [OptimizationMethod.QUANTIZATION]
        output_model.precision = [ModelPrecision.INT8]

        self.task_environment.model = output_model
        self.inferencer = self.load_inferencer()

        if optimization_parameters is not None:
            optimization_parameters.update_progress(100, None)
        logger.info("POT optimization completed")

    def load_inferencer(self) -> OpenVINOInferencer:
        """Create the OpenVINO inferencer object.

        Returns:
            OpenVINOInferencer object
        """
        if self.task_environment.model is None:
            raise Exception("task_environment.model is None. Cannot load weights.")
        return OpenVINOInferencer(
            config=OmegaConf.create(self.config.to_dict()),
            path=(
                self.task_environment.model.get_data("openvino.xml"),
                self.task_environment.model.get_data("openvino.bin"),
            ),
        )

    @staticmethod
    def __save_weights(path: str, data: bytes) -> None:
        """Write data to file.

        Args:
            path (str): Path of output file
            data (bytes): Data to write
        """
        with open(path, "wb") as file:
            file.write(data)

    @staticmethod
    def __load_weights(path: str, output_model: ModelEntity, key: str) -> None:
        """Load weights into output model.

        Args:
            path (str): Path to weights
            output_model (ModelEntity): Model to which the weights are assigned
            key (str): Key of the output model into which the weights are assigned
        """
        with open(path, "rb") as file:
            output_model.set_data(key, file.read())

    def _get_openvino_configuration(self) -> Dict[str, Any]:
        """Return configuration required by the exported model."""
        if self.task_environment.model is None:
            raise Exception("task_environment.model is None. Cannot get configuration.")

        configuration = {
            "image_threshold": np.frombuffer(
                self.task_environment.model.get_data("image_threshold"), dtype=np.float32
            ).item(),
            "min": np.frombuffer(self.task_environment.model.get_data("min"), dtype=np.float32).item(),
            "max": np.frombuffer(self.task_environment.model.get_data("max"), dtype=np.float32).item(),
            "labels": LabelSchemaMapper.forward(self.task_environment.label_schema),
            "threshold": 0.5,
        }
        if "transforms" not in self.config.keys():
            configuration["mean_values"] = list(np.array([0.485, 0.456, 0.406]) * 255)
            configuration["scale_values"] = list(np.array([0.229, 0.224, 0.225]) * 255)
        else:
            configuration["mean_values"] = self.config.transforms.mean
            configuration["scale_values"] = self.config.transforms.std

        if "pixel_threshold" in self.task_environment.model.model_adapters.keys():
            configuration["pixel_threshold"] = np.frombuffer(
                self.task_environment.model.get_data("pixel_threshold"), dtype=np.float32
            ).item()
        else:
            configuration["pixel_threshold"] = np.frombuffer(
                self.task_environment.model.get_data("image_threshold"), dtype=np.float32
            ).item()

        return configuration

    def deploy(self, output_model: ModelEntity) -> None:
        """Exports the weights from ``output_model`` along with exportable code.

        Args:
            output_model (ModelEntity): Model with ``openvino.xml`` and ``.bin`` keys

        Raises:
            Exception: If ``task_environment.model`` is None
        """
        logger.info("Deploying Model")

        if self.task_environment.model is None:
            raise Exception("task_environment.model is None. Cannot load weights.")

        work_dir = os.path.dirname(demo.__file__)
        parameters: Dict[str, Any] = {}

        task_type = str(self.task_type).lower()

        parameters["type_of_model"] = task_type
        parameters["converter_type"] = task_type.upper()
        parameters["model_parameters"] = self._get_openvino_configuration()
        zip_buffer = io.BytesIO()
        with ZipFile(zip_buffer, "w") as arch:
            # model files
            arch.writestr(os.path.join("model", "model.xml"), self.task_environment.model.get_data("openvino.xml"))
            arch.writestr(os.path.join("model", "model.bin"), self.task_environment.model.get_data("openvino.bin"))
            arch.writestr(os.path.join("model", "config.json"), json.dumps(parameters, ensure_ascii=False, indent=4))
            # model_wrappers files
            for root, _, files in os.walk(
                os.path.dirname(otx.algorithms.anomaly.adapters.anomalib.exportable_code.__file__)
            ):
                for file in files:
                    file_path = os.path.join(root, file)
                    arch.write(
                        file_path, os.path.join("python", "model_wrappers", file_path.split("exportable_code/")[1])
                    )
            # other python files
            arch.write(os.path.join(work_dir, "requirements.txt"), os.path.join("python", "requirements.txt"))
            arch.write(os.path.join(work_dir, "LICENSE"), os.path.join("python", "LICENSE"))
            arch.write(os.path.join(work_dir, "README.md"), os.path.join("python", "README.md"))
            arch.write(os.path.join(work_dir, "demo.py"), os.path.join("python", "demo.py"))
        output_model.exportable_code = zip_buffer.getvalue()
        logger.info("Deployment completed.")



import math

import numpy as np
from torch.utils.data.sampler import Sampler

from otx.mpa.utils.logger import get_logger

logger = get_logger()


class BalancedSampler(Sampler):
    """Sampler for Class-Incremental Task
    This sampler is a sampler that creates an effective batch
    In reduce mode,
    reduce the iteration size by estimating the trials
    that all samples in the tail class are selected more than once with probability 0.999

    Args:
        dataset (Dataset): A built-up dataset
        samples_per_gpu (int): batch size of Sampling
        efficient_mode (bool): Flag about using efficient mode
    """

    def __init__(self, dataset, batch_size, efficient_mode=True, num_replicas=1, rank=0, drop_last=False):
        self.batch_size = batch_size
        self.repeat = 1
        if hasattr(dataset, "times"):
            self.repeat = dataset.times
        if hasattr(dataset, "dataset"):
            self.dataset = dataset.dataset
        else:
            self.dataset = dataset
        self.img_indices = self.dataset.img_indices
        self.num_cls = len(self.img_indices.keys())
        self.data_length = len(self.dataset)
        self.num_replicas = num_replicas
        self.rank = rank
        self.drop_last = drop_last

        if efficient_mode:
            # Reduce the # of sampling (sampling data for a single epoch)
            self.num_tail = min([len(cls_indices) for cls_indices in self.img_indices.values()])
            base = 1 - (1 / self.num_tail)
            if base == 0:
                raise ValueError("Required more than one sample per class")
            self.num_trials = int(math.log(0.001, base))
            if int(self.data_length / self.num_cls) < self.num_trials:
                self.num_trials = int(self.data_length / self.num_cls)
        else:
            self.num_trials = int(self.data_length / self.num_cls)
        self.num_samples = self._calculate_num_samples()

        logger.info(f"This sampler will select balanced samples {self.num_trials} times")

    def _calculate_num_samples(self):
        num_samples = self.num_trials * self.num_cls * self.repeat

        if self.num_replicas > 1:
            # If the dataset length is evenly divisible by # of replicas, then there
            # is no need to drop any data, since the dataset will be split equally.
            if self.drop_last and num_samples % self.num_replicas != 0:  # type: ignore
                # Split to nearest available length that is evenly divisible.
                # This is to ensure each rank receives the same amount of data when
                # using this Sampler.
                num_samples = math.ceil(
                    # `type:ignore` is required because Dataset cannot provide a default __len__
                    # see NOTE in pytorch/torch/utils/data/sampler.py
                    (num_samples - self.num_replicas)
                    / self.num_replicas  # type: ignore
                )
            else:
                num_samples = math.ceil(num_samples / self.num_replicas)  # type: ignore
            self.total_size = num_samples * self.num_replicas

        return num_samples

    def __iter__(self):
        indices = []
        for _ in range(self.repeat):
            for i in range(self.num_trials):
                indice = np.concatenate(
                    [np.random.choice(self.img_indices[cls_indices], 1) for cls_indices in self.img_indices.keys()]
                )
                indices.append(indice)

        indices = np.concatenate(indices)
        indices = indices.astype(np.int64).tolist()

        if self.num_replicas > 1:
            if not self.drop_last:
                # add extra samples to make it evenly divisible
                padding_size = self.total_size - len(indices)
                if padding_size <= len(indices):
                    indices += indices[:padding_size]
                else:
                    indices += (indices * math.ceil(padding_size / len(indices)))[:padding_size]
            else:
                # remove tail of data to make it evenly divisible.
                indices = indices[: self.total_size]
            assert len(indices) == self.total_size

            # split and distribute indices
            len_indices = len(indices)
            indices = indices[
                self.rank * len_indices // self.num_replicas : (self.rank + 1) * len_indices // self.num_replicas
            ]

            assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self):
        return self.num_samples



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import glob
import os
import os.path as osp
import time

from mmcv import get_git_hash
from mmseg import __version__
from mmseg.apis import train_segmentor
from mmseg.datasets import build_dataset
from mmseg.utils import collect_env
from torch import nn

from otx.mpa.registry import STAGES
from otx.mpa.utils.logger import get_logger

from .stage import SegStage

logger = get_logger()


@STAGES.register_module()
class SegTrainer(SegStage):
    def run(self, model_cfg, model_ckpt, data_cfg, **kwargs):
        """Run training stage for segmentation

        - Configuration
        - Environment setup
        - Run training via MMSegmentation -> MMCV
        """
        self._init_logger()
        mode = kwargs.get("mode", "train")
        if mode not in self.mode:
            logger.warning(f"Supported modes are {self.mode} but '{mode}' is given.")
            return {}

        cfg = self.configure(model_cfg, model_ckpt, data_cfg, **kwargs)
        logger.info("train!")

        # FIXME: what is this? Why do we need?
        if cfg.runner.type == "IterBasedRunner":
            cfg.runner = dict(type=cfg.runner.type, max_iters=cfg.runner.max_iters)

        timestamp = time.strftime("%Y%m%d_%H%M%S", time.localtime())

        # Environment
        logger.info(f"cfg.gpu_ids = {cfg.gpu_ids}, distributed = {self.distributed}")
        env_info_dict = collect_env()
        env_info = "\n".join([(f"{k}: {v}") for k, v in env_info_dict.items()])
        dash_line = "-" * 60 + "\n"
        logger.info("Environment info:\n" + dash_line + env_info + "\n" + dash_line)

        # Data
        datasets = [build_dataset(cfg.data.train)]

        # FIXME: Currently segmentor does not support multi batch evaluation.
        # For the Self-SL case, there is no val data. So, need to check the
        if "val" in cfg.data and "val_dataloader" in cfg.data:
            cfg.data.val_dataloader.samples_per_gpu = 1

        # Target classes
        if "task_adapt" in cfg:
            target_classes = cfg.task_adapt.final
        else:
            target_classes = datasets[0].CLASSES

        # Metadata
        meta = dict()
        meta["env_info"] = env_info
        meta["seed"] = cfg.seed
        meta["exp_name"] = cfg.work_dir
        if cfg.checkpoint_config is not None:
            cfg.checkpoint_config.meta = dict(
                mmseg_version=__version__ + get_git_hash()[:7],
                CLASSES=target_classes,
            )

        self.configure_samples_per_gpu(cfg, "train", self.distributed)
        self.configure_fp16_optimizer(cfg, self.distributed)

        # Model
        model_builder = kwargs.get("model_builder", None)
        model = self.build_model(cfg, model_builder)
        model.train()
        model.CLASSES = target_classes

        if self.distributed:
            self._modify_cfg_for_distributed(model, cfg)

        self.configure_compat_cfg(cfg)

        # Save config
        # cfg.dump(osp.join(cfg.work_dir, 'config.py'))
        # logger.info(f'Config:\n{cfg.pretty_text}')

        validate = True if cfg.data.get("val", None) else False
        train_segmentor(
            model,
            datasets,
            cfg,
            distributed=self.distributed,
            validate=validate,
            timestamp=timestamp,
            meta=meta,
        )

        # Save outputs
        output_ckpt_path = os.path.join(cfg.work_dir, "latest.pth")
        best_ckpt_path = glob.glob(os.path.join(cfg.work_dir, "best_mDice_*.pth"))
        if len(best_ckpt_path) > 0:
            output_ckpt_path = best_ckpt_path[0]
        best_ckpt_path = glob.glob(os.path.join(cfg.work_dir, "best_mIoU_*.pth"))
        if len(best_ckpt_path) > 0:
            output_ckpt_path = best_ckpt_path[0]
        return dict(
            final_ckpt=output_ckpt_path,
        )

    def _modify_cfg_for_distributed(self, model, cfg):
        nn.SyncBatchNorm.convert_sync_batchnorm(model)

        if cfg.dist_params.get("linear_scale_lr", False):
            new_lr = len(cfg.gpu_ids) * cfg.optimizer.lr
            logger.info(
                f"enabled linear scaling rule to the learning rate. \
                changed LR from {cfg.optimizer.lr} to {new_lr}"
            )
            cfg.optimizer.lr = new_lr



"""MobileNet-V3-large-1 config for semi-supervised multi-class classification."""

# pylint: disable=invalid-name

_base_ = "../../base/models/mobilenet_v3.py"

model = dict(
    type="SemiSLClassifier",
    task="classification",
    backbone=dict(mode="large"),
    head=dict(
        type="SemiNonLinearClsHead",
        in_channels=960,
        hid_channels=1280,
        loss=dict(
            type="CrossEntropyLoss",
            loss_weight=1.0,
        ),
    ),
)

fp16 = dict(loss_scale=512.0)



"""Tests for OTX CLI commands"""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
import sys
from unittest.mock import patch

import pytest

from otx.cli.tools import cli
from tests.test_suite.e2e_test_system import e2e_pytest_component
from tests.test_suite.run_test_command import (
    otx_build_auto_config,
    otx_build_backbone_testing,
    otx_build_testing,
    otx_find_testing,
    otx_train_auto_config,
)

otx_dir = os.getcwd()


build_backbone_args = [
    ("CLASSIFICATION", "torchvision.mobilenet_v3_large"),
    ("CLASSIFICATION", "mmcls.MMOVBackbone"),
    ("DETECTION", "torchvision.mobilenet_v3_large"),
    ("INSTANCE_SEGMENTATION", "torchvision.mobilenet_v3_large"),
    ("SEGMENTATION", "torchvision.mobilenet_v3_large"),
]
build_backbone_args_ids = [f"{task}_{backbone}" for task, backbone in build_backbone_args]

rebuild_args = {
    "classification": {
        "default": "EfficientNet-B0",
        "--task": "classification",
        "--model": "MobileNet-V3-large-1x",
        "--train-type": "semisupervised",
    },
    "detection": {"default": "ATSS", "--task": "detection", "--model": "SSD", "--train-type": "semisupervised"},
}


class TestToolsOTXCLI:
    @e2e_pytest_component
    def test_otx_find(self):
        otx_find_testing()

    @e2e_pytest_component
    @pytest.mark.parametrize("build_backbone_args", build_backbone_args, ids=build_backbone_args_ids)
    def test_otx_backbone_build(self, tmp_dir_path, build_backbone_args):
        tmp_dir_path = tmp_dir_path / build_backbone_args[0] / build_backbone_args[1]
        otx_build_backbone_testing(tmp_dir_path, build_backbone_args)

    @e2e_pytest_component
    @pytest.mark.parametrize("case", rebuild_args.keys())
    def test_otx_build_rebuild(self, tmp_dir_path, case):
        tmp_dir_path = tmp_dir_path / "test_rebuild" / case
        # 1. Only Task
        build_arg = {"--task": rebuild_args[case]["--task"]}
        expected = {"model": rebuild_args[case]["default"], "train_type": "INCREMENTAL"}
        otx_build_testing(tmp_dir_path, build_arg, expected=expected)
        # 2. Change Model
        build_arg = {"--model": rebuild_args[case]["--model"]}
        expected = {"model": rebuild_args[case]["--model"], "train_type": "INCREMENTAL"}
        otx_build_testing(tmp_dir_path, build_arg, expected=expected)
        # 3. Change Train-type
        build_arg = {"--train-type": rebuild_args[case]["--train-type"]}
        expected = {"model": rebuild_args[case]["--model"], "train_type": rebuild_args[case]["--train-type"]}
        otx_build_testing(tmp_dir_path, build_arg, expected=expected)
        # 4. Change to Default
        build_arg = {"--model": rebuild_args[case]["default"], "--train-type": "INCREMENTAL"}
        expected = {"model": rebuild_args[case]["default"], "train_type": "INCREMENTAL"}
        otx_build_testing(tmp_dir_path, build_arg, expected=expected)


build_auto_config_args = {
    "classification": {"--train-data-roots": "tests/assets/imagenet_dataset"},
    "classification_with_task": {"--task": "classification", "--train-data-roots": "tests/assets/imagenet_dataset"},
    "detection": {"--train-data-roots": "tests/assets/car_tree_bug"},
    "detection_with_task": {"--task": "detection", "--train-data-roots": "tests/assets/car_tree_bug"},
}


class TestToolsOTXBuildAutoConfig:
    @e2e_pytest_component
    @pytest.mark.parametrize("case", build_auto_config_args.keys())
    def test_otx_build_with_autosplit(self, case, tmp_dir_path):
        otx_dir = os.getcwd()
        tmp_dir_path = tmp_dir_path / "test_build_auto_config" / case
        otx_build_auto_config(root=tmp_dir_path, otx_dir=otx_dir, args=build_auto_config_args[case])


train_auto_config_args = {
    "classification": {"--train-data-roots": "tests/assets/imagenet_dataset"},
    "classification_with_template": {
        "template": "otx/algorithms/classification/configs/mobilenet_v3_large_1_cls_incr/template.yaml",
        "--train-data-roots": "tests/assets/imagenet_dataset",
    },
    "detection": {"--train-data-roots": "tests/assets/car_tree_bug"},
    "detection_with_template": {
        "template": "otx/algorithms/detection/configs/detection/mobilenetv2_ssd/template.yaml",
        "--train-data-roots": "tests/assets/car_tree_bug",
    },
}

train_params = [
    "params",
    "--learning_parameters.num_iters",
    "1",
]


class TestToolsOTXTrainAutoConfig:
    @e2e_pytest_component
    @pytest.mark.parametrize("case", train_auto_config_args.keys())
    def test_otx_train(self, case, tmp_dir_path):
        otx_dir = os.getcwd()
        tmp_dir_path = tmp_dir_path / "test_train_auto_config" / case
        train_auto_config_args[case]["train_params"] = train_params
        otx_train_auto_config(root=tmp_dir_path, otx_dir=otx_dir, args=train_auto_config_args[case])


class TestTelemetryIntegration:
    _CMDS = ["demo", "build", "deploy", "eval", "explain", "export", "find", "optimize", "train"]

    @e2e_pytest_component
    @patch("otx.cli.utils.telemetry.init_telemetry_session", return_value=None)
    @patch("otx.cli.utils.telemetry.close_telemetry_session", return_value=None)
    @patch("otx.cli.utils.telemetry.send_version", return_value=None)
    @patch("otx.cli.utils.telemetry.send_cmd_results", return_value=None)
    def test_tm_integration_exit_0(
        self,
        mock_send_cmd,
        mock_send_version,
        mock_close_tm,
        mock_init_tm,
    ):
        backup_argv = sys.argv
        for cmd in self._CMDS:
            sys.argv = ["otx", cmd]
            with patch(f"otx.cli.tools.cli.otx_{cmd}", return_value=None) as mock_cmd:
                ret = cli.main()

            assert ret == 0
            mock_cmd.assert_called_once()
            mock_init_tm.assert_called_once()
            mock_close_tm.assert_called_once()
            mock_send_cmd.assert_called_once_with(None, cmd, {"retcode": 0})
            # reset mock state
            mock_init_tm.reset_mock()
            mock_close_tm.reset_mock()
            mock_send_cmd.reset_mock()
        sys.argv = backup_argv

    @e2e_pytest_component
    @patch("otx.cli.utils.telemetry.init_telemetry_session", return_value=None)
    @patch("otx.cli.utils.telemetry.close_telemetry_session", return_value=None)
    @patch("otx.cli.utils.telemetry.send_version", return_value=None)
    @patch("otx.cli.utils.telemetry.send_cmd_results", return_value=None)
    def test_tm_integration_exit_exception(
        self,
        mock_send_cmd,
        mock_send_version,
        mock_close_tm,
        mock_init_tm,
    ):
        backup_argv = sys.argv
        for cmd in self._CMDS:
            with patch(f"otx.cli.tools.cli.otx_{cmd}", side_effect=Exception()):
                sys.argv = ["otx", cmd]
                with pytest.raises(Exception) as e:
                    cli.main()

            assert e.type == Exception, f"{e}"
            mock_init_tm.assert_called_once()
            mock_close_tm.assert_called_once()
            mock_send_cmd.assert_called_once_with(None, cmd, {"retcode": -1, "exception": repr(Exception())})
            # reset mock state
            mock_init_tm.reset_mock()
            mock_close_tm.reset_mock()
            mock_send_cmd.reset_mock()
        sys.argv = backup_argv



"""Fixtures for anomaly tests."""
# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass

import pytest

from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.task_environment import TaskEnvironment
from tests.unit.algorithms.anomaly.helpers.dummy_dataset import get_shapes_dataset
from tests.unit.algorithms.anomaly.helpers.utils import create_task_environment


@dataclass(frozen=True)  # this ensures that the objects are immutable across tests
class TestEnvironment:
    """Test environment for anomaly tests."""

    task_environment: TaskEnvironment
    output_model: ModelEntity
    dataset: DatasetEntity
    task_type: TaskType


@pytest.fixture(
    scope="session", params=[TaskType.ANOMALY_CLASSIFICATION, TaskType.ANOMALY_DETECTION, TaskType.ANOMALY_SEGMENTATION]
)
def setup_task_environment(request):
    """Returns a task environment, a model and datset."""
    task_type = request.param
    dataset: DatasetEntity = get_shapes_dataset(task_type, one_each=True)
    task_environment = create_task_environment(dataset, task_type)
    output_model = ModelEntity(
        dataset,
        task_environment.get_model_configuration(),
    )
    environment = TestEnvironment(task_environment, output_model, dataset, task_type)
    return environment



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest
from openvino.model_zoo.model_api.adapters.openvino_adapter import OpenvinoAdapter
from openvino.model_zoo.model_api.models import SegmentationModel

from otx.algorithms.segmentation.adapters.openvino.model_wrappers.blur import (
    BlurSegmentation,
    get_activation_map,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_get_activation_map():
    fake_features = np.array([0.1, 0.2, 0.3, 0.4, 0.5])
    returned_value = get_activation_map(fake_features)

    assert type(returned_value).__module__ == np.__name__
    assert len(returned_value) == len(fake_features)


class TestBlurSegmentation:
    @pytest.fixture(autouse=True)
    def setup(self, mocker):
        class MockAdapter(OpenvinoAdapter):
            def __init__(self):
                pass

        mocker.patch.object(SegmentationModel, "__init__")
        model_adapter = MockAdapter()
        self.blur = BlurSegmentation(model_adapter)

    @e2e_pytest_unit
    def test_parameters(self):
        params = BlurSegmentation.parameters()

        assert "blur_strength" in params
        assert "soft_threshold" in params

    @e2e_pytest_unit
    def test_get_outputs(self):
        self.blur.outputs = {"output": np.ones((2, 3, 4))}
        returned_value = self.blur._get_outputs()

        assert returned_value == "output"

    @e2e_pytest_unit
    def test_postprocess(self, mocker):
        self.blur.output_blob_name = "output"
        self.blur.soft_threshold = 0.5
        self.blur.blur_strength = 2
        fake_output = {"output": np.ones((2, 3, 4))}
        fake_metadata = {"original_shape": (2, 3, 4)}
        returned_value = self.blur.postprocess(outputs=fake_output, meta=fake_metadata)

        assert type(returned_value).__module__ == np.__name__
        assert fake_metadata["feature_vector"] is None
        assert fake_metadata["soft_prediction"] is not None



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import functools

import torch
from mmdet.models.builder import DETECTORS
from mmdet.models.detectors.single_stage import SingleStageDetector

from otx.mpa.deploy.utils import is_mmdeploy_enabled
from otx.mpa.modules.utils.task_adapt import map_class_names
from otx.mpa.utils.logger import get_logger

from .l2sp_detector_mixin import L2SPDetectorMixin
from .sam_detector_mixin import SAMDetectorMixin

logger = get_logger()


@DETECTORS.register_module()
class CustomSingleStageDetector(SAMDetectorMixin, L2SPDetectorMixin, SingleStageDetector):
    """SAM optimizer & L2SP regularizer enabled custom SSD"""

    def __init__(self, *args, task_adapt=None, **kwargs):
        super().__init__(*args, **kwargs)

        # Hook for class-sensitive weight loading
        if task_adapt:
            self._register_load_state_dict_pre_hook(
                functools.partial(
                    self.load_state_dict_pre_hook,
                    self,  # model
                    task_adapt["dst_classes"],  # model_classes
                    task_adapt["src_classes"],  # chkpt_classes
                )
            )

    def forward_train(self, img, img_metas, gt_bboxes, gt_labels, gt_bboxes_ignore=None, **kwargs):
        """
        Args:
            img (Tensor): Input images of shape (N, C, H, W).
                Typically these should be mean centered and std scaled.
            img_metas (list[dict]): A List of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.
                For details on the values of these keys see
                :class:`mmdet.datasets.pipelines.Collect`.
            gt_bboxes (list[Tensor]): Each item are the truth boxes for each
                image in [tl_x, tl_y, br_x, br_y] format.
            gt_labels (list[Tensor]): Class indices corresponding to each box
            gt_bboxes_ignore (None | list[Tensor]): Specify which bounding
                boxes can be ignored when computing the loss.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        batch_input_shape = tuple(img[0].size()[-2:])
        for img_meta in img_metas:
            img_meta["batch_input_shape"] = batch_input_shape
        x = self.extract_feat(img)
        losses = self.bbox_head.forward_train(x, img_metas, gt_bboxes, gt_labels, gt_bboxes_ignore, **kwargs)
        return losses

    @staticmethod
    def load_state_dict_pre_hook(model, model_classes, chkpt_classes, chkpt_dict, prefix, *args, **kwargs):
        """Modify input state_dict according to class name matching before weight loading"""
        logger.info(f"----------------- CustomSSD.load_state_dict_pre_hook() called w/ prefix: {prefix}")

        # Dst to chkpt mapping index (including BG class)
        model_dict = model.state_dict()
        chkpt_classes = list(chkpt_classes) + ["__BG__"]
        model_classes = list(model_classes) + ["__BG__"]
        num_chkpt_classes = len(chkpt_classes)
        num_model_classes = len(model_classes)
        model2chkpt = map_class_names(model_classes, chkpt_classes)
        logger.info(f"{chkpt_classes} -> {model_classes}")

        # List of class-relevant params
        if prefix + "bbox_head.cls_convs.0.weight" in chkpt_dict:
            param_names = [
                "bbox_head.cls_convs.{}.weight",  # normal
                "bbox_head.cls_convs.{}.bias",
            ]
        elif prefix + "bbox_head.cls_convs.0.0.weight" in chkpt_dict:
            param_names = [
                "bbox_head.cls_convs.{}.3.weight",  # depth-wise: (0)conv -> (1)bn -> (2)act -> (3)conv
                "bbox_head.cls_convs.{}.3.bias",
            ]
        else:
            param_names = []

        # Weight mixing
        for level in range(10):  # For each level (safer inspection loop than 'while True'. Mostly has 2~3 levels)
            level_found = False
            for model_name in param_names:
                model_name = model_name.format(level)
                chkpt_name = prefix + model_name
                if model_name not in model_dict or chkpt_name not in chkpt_dict:
                    logger.info(f"Skipping weight copy: {chkpt_name}")
                    break
                level_found = True

                model_param = model_dict[model_name].clone()
                chkpt_param = chkpt_dict[chkpt_name]

                num_chkpt_anchors = int(chkpt_param.shape[0] / num_chkpt_classes)
                num_model_anchors = int(model_param.shape[0] / num_model_classes)
                num_anchors = min(num_chkpt_anchors, num_model_anchors)
                logger.info(
                    f"Mixing {model_name}: {num_chkpt_anchors}x{num_chkpt_classes} -> "
                    f"{num_model_anchors}x{num_model_classes} anchors"
                )

                for anchor_idx in range(num_anchors):  # For each anchor
                    for m, c in enumerate(model2chkpt):
                        if c >= 0:
                            # Copying only matched weight rows
                            model_param[anchor_idx * num_model_classes + m].copy_(
                                chkpt_param[anchor_idx * num_chkpt_classes + c]
                            )

                # Replace checkpoint weight by mixed weights
                chkpt_dict[chkpt_name] = model_param
            if not level_found:
                break


if is_mmdeploy_enabled():
    from mmdeploy.core import FUNCTION_REWRITER, mark
    from mmdeploy.utils import is_dynamic_shape

    from otx.mpa.modules.hooks.recording_forward_hooks import (
        DetSaliencyMapHook,
        FeatureVectorHook,
    )

    @FUNCTION_REWRITER.register_rewriter(
        "otx.mpa.modules.models.detectors.custom_single_stage_detector.CustomSingleStageDetector.simple_test"
    )
    def custom_single_stage_detector__simple_test(ctx, self, img, img_metas, **kwargs):
        feat = self.extract_feat(img)
        outs = self.bbox_head(feat)
        bbox_results = self.bbox_head.get_bboxes(*outs, img_metas=img_metas, cfg=self.test_cfg, **kwargs)

        if ctx.cfg["dump_features"]:
            feature_vector = FeatureVectorHook.func(feat)
            cls_scores = outs[0]
            saliency_map = DetSaliencyMapHook(self).func(cls_scores, cls_scores_provided=True)
            return (*bbox_results, feature_vector, saliency_map)

        return bbox_results

    @mark("custom_ssd_forward", inputs=["input"], outputs=["dets", "labels", "feats", "saliencies"])
    def __forward_impl(ctx, self, img, img_metas, **kwargs):
        assert isinstance(img, torch.Tensor)

        deploy_cfg = ctx.cfg
        is_dynamic_flag = is_dynamic_shape(deploy_cfg)
        # get origin input shape as tensor to support onnx dynamic shape
        img_shape = torch._shape_as_tensor(img)[2:]
        if not is_dynamic_flag:
            img_shape = [int(val) for val in img_shape]
        img_metas[0]["img_shape"] = img_shape
        return self.simple_test(img, img_metas, **kwargs)

    @FUNCTION_REWRITER.register_rewriter(
        "otx.mpa.modules.models.detectors.custom_single_stage_detector.CustomSingleStageDetector.forward"
    )
    def custom_ssd__forward(ctx, self, img, img_metas=None, return_loss=False, **kwargs):
        if img_metas is None:
            img_metas = [{}]
        else:
            assert len(img_metas) == 1, "do not support aug_test"
            img_metas = img_metas[0]

        if isinstance(img, list):
            img = img[0]

        return __forward_impl(ctx, self, img, img_metas=img_metas, **kwargs)



"""Openvino Task of OTX Action Recognition."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
import json
import logging
import os
import tempfile
from typing import Any, Dict, List, Optional, Tuple, Union
from zipfile import ZipFile

import numpy as np
from addict import Dict as ADDict
from compression.api import DataLoader
from compression.engines.ie_engine import IEEngine
from compression.graph import load_model, save_model
from compression.graph.model_utils import compress_model_weights, get_nodes_by_type
from compression.pipeline.initializer import create_pipeline

from otx.algorithms.action.adapters.openvino import (
    ActionOVClsDataLoader,
    get_ovdataloader,
    model_wrappers,
)
from otx.algorithms.action.configs.base import ActionConfig
from otx.api.entities.annotation import AnnotationSceneEntity
from otx.api.entities.datasets import DatasetEntity, DatasetItemEntity
from otx.api.entities.inference_parameters import (
    InferenceParameters,
    default_progress_callback,
)
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
    OptimizationMethod,
)
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.serialization.label_mapper import LabelSchemaMapper, label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.exportable_code import demo
from otx.api.usecases.exportable_code.inference import BaseInferencer
from otx.api.usecases.exportable_code.prediction_to_annotation_converter import (
    ClassificationToAnnotationConverter,
    DetectionBoxToAnnotationConverter,
    IPredictionToAnnotationConverter,
)
from otx.api.usecases.tasks.interfaces.deployment_interface import IDeploymentTask
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.optimization_interface import (
    IOptimizationTask,
    OptimizationType,
)
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)

try:
    from openvino.model_zoo.model_api.adapters import OpenvinoAdapter, create_core
    from openvino.model_zoo.model_api.models import Model
except ImportError:
    import warnings

    warnings.warn("ModelAPI was not found.")

logger = logging.getLogger(__name__)


# TODO: refactoring to Sphinx style.
class ActionOpenVINOInferencer(BaseInferencer):
    """ActionOpenVINOInferencer class in OpenVINO task for action recognition."""

    @check_input_parameters_type()
    def __init__(
        self,
        task_type: str,
        hparams: ActionConfig,
        label_schema: LabelSchemaEntity,
        model_file: Union[str, bytes],
        weight_file: Union[str, bytes, None] = None,
        device: str = "CPU",
        num_requests: int = 1,
    ):  # pylint: disable=unused-argument, too-many-arguments
        """Inferencer implementation for OTX Action Recognition using OpenVINO backend.

        Args:
            task_type (str): Type of action task. ["ACTION_CLASSIFICATION", "ACTION_DETECTION"]
            hparams (ActionConfig): Hyperparameters for action task
            label_schema (LabelSchemaEntity): Label schema for model file
            model_file (Union[str, bytes]): XML file for model structure
            weight_file (Union[str, bytes, None]): Model weights file
                Default: None
            device (str): Device for inference. Default: "CPU"
            num_requests (int): Number of requests
        """

        self.task_type = task_type
        self.label_schema = label_schema
        model_adapter = OpenvinoAdapter(
            create_core(), model_file, weight_file, device=device, max_num_requests=num_requests
        )
        self.configuration: Dict[Any, Any] = {}
        self.model = Model.create_model(self.task_type, model_adapter, self.configuration, preload=True)
        self.converter: IPredictionToAnnotationConverter
        if self.task_type == "ACTION_CLASSIFICATION":
            self.converter = ClassificationToAnnotationConverter(self.label_schema)
        else:
            self.converter = DetectionBoxToAnnotationConverter(self.label_schema)

    @check_input_parameters_type()
    def pre_process(self, image: List[DatasetItemEntity]) -> Tuple[Dict[str, np.ndarray], Dict[str, Any]]:
        """Pre-process function of OpenVINO Inferencer for Action Recognition."""
        return self.model.preprocess(image)

    @check_input_parameters_type()
    def post_process(
        self, prediction: Dict[str, np.ndarray], metadata: Dict[str, Any]
    ) -> Optional[AnnotationSceneEntity]:
        """Post-process function of OpenVINO Inferencer for Action Recognition."""

        prediction = self.model.postprocess(prediction, metadata)
        return self.converter.convert_to_annotation(prediction, metadata)

    @check_input_parameters_type()
    def predict(self, image: List[DatasetItemEntity]) -> AnnotationSceneEntity:
        """Predict function of OpenVINO Action Inferencer for Action Recognition."""
        data, metadata = self.pre_process(image)
        raw_predictions = self.forward(data)
        predictions = self.post_process(raw_predictions, metadata)
        return predictions

    # @check_input_parameters_type()
    def forward(self, image: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        """Forward function of OpenVINO Action Inferencer for Action Recognition."""

        return self.model.infer_sync(image)


class DataLoaderWrapper(DataLoader):
    """DataLoader implementation for ActionOpenVINOTask."""

    @check_input_parameters_type()
    def __init__(self, dataloader: DataLoader, inferencer: BaseInferencer):
        super().__init__(config=None)
        self.dataloader = dataloader
        self.inferencer = inferencer

    @check_input_parameters_type()
    def __getitem__(self, index: int):
        """Get item from dataset."""
        item = self.dataloader[index]
        annotation = item[len(item) // 2].annotation_scene
        inputs, metadata = self.inferencer.pre_process(item)
        return (index, annotation), inputs, metadata

    def __len__(self):
        """Get length of dataset."""
        return len(self.dataloader)


class ActionOpenVINOTask(IDeploymentTask, IInferenceTask, IEvaluationTask, IOptimizationTask):
    """Task implementation for OTX Action Recognition using OpenVINO backend."""

    @check_input_parameters_type()
    def __init__(self, task_environment: TaskEnvironment):
        self.task_environment = task_environment
        self.hparams = self.task_environment.get_hyper_parameters(ActionConfig)
        self.model = self.task_environment.model
        self.task_type = self.task_environment.model_template.task_type.name
        self.inferencer = self.load_inferencer()

    def load_inferencer(self) -> ActionOpenVINOInferencer:
        """load_inferencer function of OpenVINOTask for Action Recognition."""

        if self.model is None:
            raise RuntimeError("load_inferencer failed, model is None")

        return ActionOpenVINOInferencer(
            self.task_type,
            self.hparams,
            self.task_environment.label_schema,
            self.model.get_data("openvino.xml"),
            self.model.get_data("openvino.bin"),
        )

    # pylint: disable=no-value-for-parameter
    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def infer(
        self, dataset: DatasetEntity, inference_parameters: Optional[InferenceParameters] = None
    ) -> DatasetEntity:
        """Infer function of OpenVINOTask for Action Recognition."""
        update_progress_callback = default_progress_callback
        clip_len = self.inferencer.model.t
        width = self.inferencer.model.w
        height = self.inferencer.model.h
        dataloader = get_ovdataloader(dataset, self.task_type, clip_len, width, height)
        dataset_size = len(dataloader)
        for i, data in enumerate(dataloader):
            prediction = self.inferencer.predict(data)
            if isinstance(dataloader, ActionOVClsDataLoader):
                dataloader.add_prediction(dataset, data, prediction)
            else:
                dataloader.add_prediction(data, prediction)
            update_progress_callback(int(i / dataset_size * 100))
        return dataset

    @check_input_parameters_type()
    def evaluate(self, output_resultset: ResultSetEntity, evaluation_metric: Optional[str] = None):
        """Evaluate function of OpenVINOTask."""

        if evaluation_metric is not None:
            logger.warning(f"Requested to use {evaluation_metric} metric," "but parameter is ignored.")
        if self.task_type == "ACTION_CLASSIFICATION":
            output_resultset.performance = MetricsHelper.compute_accuracy(output_resultset).get_performance()
        elif self.task_type == "ACTION_DETECTION":
            output_resultset.performance = MetricsHelper.compute_f_measure(output_resultset).get_performance()

    @check_input_parameters_type()
    def deploy(self, output_model: ModelEntity) -> None:
        """Deploy function of OpenVINOTask."""

        logger.info("Deploying the model")

        work_dir = os.path.dirname(demo.__file__)
        parameters = {}  # type: Dict[Any, Any]
        parameters["type_of_model"] = f"otx_{self.task_type.lower()}"
        parameters["converter_type"] = f"{self.task_type}"
        parameters["model_parameters"] = self.inferencer.configuration
        parameters["model_parameters"]["labels"] = LabelSchemaMapper.forward(self.task_environment.label_schema)

        if self.model is None:
            raise RuntimeError("deploy failed, model is None")

        zip_buffer = io.BytesIO()
        with ZipFile(zip_buffer, "w") as arch:
            # model files
            arch.writestr(os.path.join("model", "model.xml"), self.model.get_data("openvino.xml"))
            arch.writestr(os.path.join("model", "model.bin"), self.model.get_data("openvino.bin"))
            arch.writestr(os.path.join("model", "config.json"), json.dumps(parameters, ensure_ascii=False, indent=4))
            # model_wrappers files
            for root, _, files in os.walk(os.path.dirname(model_wrappers.__file__)):
                for file in files:
                    file_path = os.path.join(root, file)
                    arch.write(
                        file_path, os.path.join("python", "model_wrappers", file_path.split("model_wrappers/")[1])
                    )
            # other python files
            arch.write(os.path.join(work_dir, "requirements.txt"), os.path.join("python", "requirements.txt"))
            arch.write(os.path.join(work_dir, "LICENSE"), os.path.join("python", "LICENSE"))
            arch.write(os.path.join(work_dir, "README.md"), os.path.join("python", "README.md"))
            arch.write(os.path.join(work_dir, "demo.py"), os.path.join("python", "demo.py"))
        output_model.exportable_code = zip_buffer.getvalue()
        logger.info("Deploying completed")

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def optimize(
        self,
        optimization_type: OptimizationType,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        optimization_parameters: Optional[OptimizationParameters] = None,
    ):  # pylint: disable=too-many-locals
        """Optimize function of OpenVINOTask."""

        if optimization_type is not OptimizationType.POT:
            raise ValueError("POT is the only supported optimization type for OpenVino models")

        clip_len = self.inferencer.model.t
        width = self.inferencer.model.w
        height = self.inferencer.model.h
        data_loader = get_ovdataloader(dataset, self.task_type, clip_len, width, height)
        data_loader = DataLoaderWrapper(data_loader, self.inferencer)

        if self.model is None:
            raise RuntimeError("optimize failed, model is None")

        with tempfile.TemporaryDirectory() as tempdir:
            xml_path = os.path.join(tempdir, "model.xml")
            bin_path = os.path.join(tempdir, "model.bin")
            with open(xml_path, "wb") as f:
                f.write(self.model.get_data("openvino.xml"))
            with open(bin_path, "wb") as f:
                f.write(self.model.get_data("openvino.bin"))

            model_config = ADDict({"model_name": "openvino_model", "model": xml_path, "weights": bin_path})

            model = load_model(model_config)

            if get_nodes_by_type(model, ["FakeQuantize"]):
                raise RuntimeError("Model is already optimized by POT")

        if optimization_parameters is not None:
            optimization_parameters.update_progress(10, None)

        engine_config = ADDict({"device": "CPU"})

        stat_subset_size = self.hparams.pot_parameters.stat_subset_size
        preset = self.hparams.pot_parameters.preset.name.lower()

        algorithms = [
            {
                "name": "DefaultQuantization",
                "params": {
                    "target_device": "ANY",
                    "preset": preset,
                    "stat_subset_size": min(stat_subset_size, len(data_loader)),
                    "shuffle_data": True,
                },
            }
        ]

        engine = IEEngine(config=engine_config, data_loader=data_loader, metric=None)

        pipeline = create_pipeline(algorithms, engine)

        compressed_model = pipeline.run(model)

        compress_model_weights(compressed_model)

        if optimization_parameters is not None:
            optimization_parameters.update_progress(90, None)

        with tempfile.TemporaryDirectory() as tempdir:
            save_model(compressed_model, tempdir, model_name="model")
            with open(os.path.join(tempdir, "model.xml"), "rb") as f:
                output_model.set_data("openvino.xml", f.read())
            with open(os.path.join(tempdir, "model.bin"), "rb") as f:
                output_model.set_data("openvino.bin", f.read())

        output_model.set_data("label_schema.json", label_schema_to_bytes(self.task_environment.label_schema))

        # set model attributes for quantized model
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.POT
        output_model.optimization_methods = [OptimizationMethod.QUANTIZATION]
        output_model.precision = [ModelPrecision.INT8]

        self.model = output_model
        self.inferencer = self.load_inferencer()

        if optimization_parameters is not None:
            optimization_parameters.update_progress(100, None)
        logger.info("POT optimization completed")



"""Model explain demonstration tool."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from pathlib import Path

from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.task_environment import TaskEnvironment
from otx.cli.manager import ConfigManager
from otx.cli.utils.importing import get_impl_class
from otx.cli.utils.io import (
    get_explain_dataset_from_filelist,
    get_image_files,
    read_label_schema,
    read_model,
    save_saliency_output,
)
from otx.cli.utils.nncf import is_checkpoint_nncf
from otx.cli.utils.parser import (
    add_hyper_parameters_sub_parser,
    get_parser_and_hprams_data,
)

ESC_BUTTON = 27
SUPPORTED_EXPLAIN_ALGORITHMS = ["activationmap", "eigencam", "classwisesaliencymap"]

# pylint: disable=too-many-locals


def get_args():
    """Parses command line arguments."""
    parser, hyper_parameters, params = get_parser_and_hprams_data()

    parser.add_argument(
        "--explain-data-roots",
        required=True,
        help="Comma-separated paths to explain data folders.",
    )
    parser.add_argument(
        "--save-explanation-to",
        default="saliency_dump",
        help="Output path for explanation images.",
    )
    parser.add_argument(
        "--load-weights",
        required=True,
        help="Load model weights from previously saved checkpoint.",
    )
    parser.add_argument(
        "--explain-algorithm",
        default="ClassWiseSaliencyMap",
        help=f"Explain algorithm name, currently support {SUPPORTED_EXPLAIN_ALGORITHMS}."
        "For Openvino task, default method will be selected.",
    )
    parser.add_argument(
        # "-w",
        "--overlay-weight",
        type=float,
        default=0.5,
        help="weight of the saliency map when overlaying the saliency map",
    )
    add_hyper_parameters_sub_parser(parser, hyper_parameters, modes=("INFERENCE",))
    override_param = [f"params.{param[2:].split('=')[0]}" for param in params if param.startswith("--")]

    return parser.parse_args(), override_param


def main():
    """Main function that is used for model explanation."""

    args, override_param = get_args()

    config_manager = ConfigManager(args, mode="eval")
    # Auto-Configuration for model template
    config_manager.configure_template()

    # Update Hyper Parameter Configs
    hyper_parameters = config_manager.get_hyparams_config(override_param)

    # Get classes for Task, ConfigurableParameters and Dataset.
    template = config_manager.template
    if any(args.load_weights.endswith(x) for x in (".bin", ".xml", ".zip")):
        task_class = get_impl_class(template.entrypoints.openvino)
    elif args.load_weights.endswith(".pth"):
        if is_checkpoint_nncf(args.load_weights):
            task_class = get_impl_class(template.entrypoints.nncf)
        else:
            task_class = get_impl_class(template.entrypoints.base)
    else:
        raise ValueError(f"Unsupported file: {args.load_weights}")

    environment = TaskEnvironment(
        model=None,
        hyper_parameters=hyper_parameters,
        label_schema=read_label_schema(args.load_weights),
        model_template=template,
    )

    environment.model = read_model(environment.get_model_configuration(), args.load_weights, None)
    task = task_class(task_environment=environment)

    if args.explain_algorithm.lower() not in SUPPORTED_EXPLAIN_ALGORITHMS:
        raise NotImplementedError(
            f"{args.explain_algorithm} currently not supported. \
            Currently only support {SUPPORTED_EXPLAIN_ALGORITHMS}"
        )
    if not Path(args.save_explanation_to).exists():
        Path(args.save_explanation_to).mkdir(parents=True)

    image_files = get_image_files(args.explain_data_roots)
    dataset_to_explain = get_explain_dataset_from_filelist(image_files)
    explain_parameters = InferenceParameters(
        is_evaluation=False,
        explainer=args.explain_algorithm,
        explain_predicted_classes=False,
    )
    explained_dataset = task.explain(
        dataset_to_explain.with_empty_annotations(),
        explain_parameters,
    )

    for explained_data, (_, filename) in zip(explained_dataset, image_files):
        for metadata in explained_data.get_metadata():
            saliency_data = metadata.data
            fname = f"{Path(Path(filename).name).stem}_{saliency_data.name}".replace(" ", "_")
            save_saliency_output(
                process_saliency_maps=explain_parameters.process_saliency_maps,
                img=explained_data.numpy,
                saliency_map=saliency_data.numpy,
                save_dir=args.save_explanation_to,
                fname=fname,
                weight=args.overlay_weight,
            )

    print(f"Saliency maps saved to {args.save_explanation_to} for {len(image_files)} images...")

    return dict(retcode=0, template=template.name)


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from copy import deepcopy
from typing import Dict, List, Optional, Union

import openvino.runtime as ov
import torch
from mmdet.models.builder import HEADS
from mmdet.models.roi_heads.bbox_heads.bbox_head import BBoxHead

from ....mmov_model import MMOVModel


@HEADS.register_module()
class MMOVBBoxHead(BBoxHead):
    def __init__(
        self,
        model_path_or_model: Union[str, ov.Model],
        weight_path: Optional[str] = None,
        inputs: Dict[str, Union[str, List[str]]] = {},
        outputs: Dict[str, Union[str, List[str]]] = {},
        init_weight: bool = False,
        verify_shape: bool = True,
        background_index: Optional[int] = None,
        *args,
        **kwargs,
    ):
        self._model_path_or_model = model_path_or_model
        self._weight_path = weight_path
        self._inputs = deepcopy(inputs)
        self._outputs = deepcopy(outputs)
        self._init_weight = init_weight
        self._verify_sahpe = verify_shape
        self._background_index = background_index
        super().__init__(*args, **kwargs)

        if self._background_index is not None and self._background_index < 0:
            self._background_index = self.num_classes + 1 - self._background_index

        if "extractor" in inputs and "extractor" in outputs:
            self.extractor = MMOVModel(
                self._model_path_or_model,
                inputs=inputs["extractor"],
                outputs=outputs["extractor"],
                remove_normalize=False,
                merge_bn=True,
                paired_bn=True,
                init_weight=self._init_weight,
                verify_shape=self._verify_sahpe,
            )

        if self.with_cls:
            assert "fc_cls" in inputs and "fc_cls" in outputs
            self.fc_cls = MMOVModel(
                self._model_path_or_model,
                inputs=inputs["fc_cls"],
                outputs=outputs["fc_cls"],
                remove_normalize=False,
                merge_bn=False,
                paired_bn=False,
                init_weight=self._init_weight,
                verify_shape=self._verify_sahpe,
            )

        if self.with_reg:
            assert "fc_reg" in inputs and "fc_reg" in outputs
            self.fc_reg = MMOVModel(
                self._model_path_or_model,
                inputs=inputs["fc_reg"],
                outputs=outputs["fc_reg"],
                remove_normalize=False,
                merge_bn=False,
                paired_bn=False,
                init_weight=self._init_weight,
                verify_shape=self._verify_sahpe,
            )

    def init_weights(self):
        # TODO
        pass

    def forward(self, x):
        if getattr(self, "extractor"):
            x = self.extractor(x)
        cls_score = self.fc_cls(x) if self.with_cls else None
        bbox_pred = self.fc_reg(x) if self.with_reg else None

        # since mmdet v2.0, BBoxHead is supposed to be
        # that FG labels to [0, num_class-1] and BG labels to num_class
        # but faster_rcnn_resnet50_coco, etc. from OMZ are
        # that FG labels to be [1, num_class] and BG labels to be 0
        if (
            self._background_index is not None
            and cls_score is not None
            and self._background_index != cls_score.shape(-1)
        ):
            cls_score = torch.cat(
                (
                    cls_score[:, : self._background_index],
                    cls_score[:, self._background_index + 1 :],
                    cls_score[:, self._background_index : self._background_index + 1],
                ),
                -1,
            )

        return (cls_score, bbox_pred)



"""Segmentation models dynamic deploy config."""

_base_ = ["./base_segmentation_static.py"]

ir_config = dict(
    dynamic_axes={
        "input": {
            0: "batch",
            2: "height",
            3: "width",
        },
        "output": {
            0: "batch",
            2: "height",
            3: "width",
        },
    },
)



"""Sample Code of otx training for detection."""

# Copyright (C) 2021-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import argparse
import sys

import cv2
import numpy as np
from mmcv.utils import get_logger

from otx.algorithms.common.utils import get_task_class
from otx.api.configuration.helper import create
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType

logger = get_logger(name="mmseg")


def parse_args():
    """Parse function for getting model template & check export."""
    parser = argparse.ArgumentParser(description="Sample showcasing the new API")
    parser.add_argument("template_file_path", help="path to template file")
    parser.add_argument("--export", action="store_true")
    return parser.parse_args()


colors = [(128, 0, 0), (0, 128, 0)]


def load_test_dataset(data_type):
    """Load Sample dataset for detection."""

    def gen_circle_image(resolution):
        width, height = resolution
        image = np.full([height, width, 3], fill_value=128, dtype=np.uint8)
        true_labels = np.full([height, width, 1], fill_value=0, dtype=np.uint8)
        cv2.circle(image, (int(height / 2), int(width / 2)), 90, (0, 0, 255), -1)
        cv2.circle(true_labels, (int(height / 2), int(width / 2)), 90, 1, -1)
        return (image, true_labels)

    def gen_rect_image(resolution):
        width, height = resolution
        image = np.full([height, width, 3], fill_value=128, dtype=np.uint8)
        true_labels = np.full([height, width, 1], fill_value=0, dtype=np.uint8)
        cv2.rectangle(image, (int(height * 0.1), int(width * 0.1)), (int(height / 2), int(width / 2)), (0, 255, 0), -1)
        cv2.rectangle(true_labels, (int(height * 0.1), int(width * 0.1)), (int(height / 2), int(width / 2)), 2, -1)
        return (image, true_labels)

    labels = [
        LabelEntity(name="circle", domain=Domain.SEGMENTATION, id=1),  # OLD class
        LabelEntity(name="rect", domain=Domain.SEGMENTATION, id=2),
    ]

    def get_image(data_type, subset, label_id):
        ignored_labels = []
        if label_id == 1:
            image, true_label = gen_circle_image((640, 480))
            if data_type == "new" and subset == Subset.TRAINING:
                ignored_labels = [LabelEntity(name="rect", domain=Domain.SEGMENTATION, id=2)]
        else:
            image, true_label = gen_rect_image((640, 480))

        height, width = true_label.shape[:2]
        label_mask = true_label == label_id
        label_index_map = label_mask.astype(np.uint8)
        contours, hierarchies = cv2.findContours(label_index_map, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
        for contour, hierarchy in zip(contours, hierarchies[0]):
            if hierarchy[3] != -1:
                continue

            contour = list(contour)
            if len(contour) <= 2:
                continue

            points = [Point(x=point[0][0] / width, y=point[0][1] / height) for point in contour]
        return DatasetItemEntity(
            media=Image(data=image),
            annotation_scene=AnnotationSceneEntity(
                annotations=[Annotation(Polygon(points=points), labels=[ScoredLabel(label=labels[label_id - 1])])],
                kind=AnnotationSceneKind.ANNOTATION,
            ),
            subset=subset,
            ignored_labels=ignored_labels,
        )

    old_train = [
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
        get_image("old", Subset.TRAINING, 1),
    ]

    old_val = [
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
        get_image("old", Subset.VALIDATION, 1),
    ]

    new_train = [
        get_image("new", Subset.TRAINING, 1),
        get_image("new", Subset.TRAINING, 1),
        get_image("new", Subset.TRAINING, 1),
        get_image("new", Subset.TRAINING, 1),
        get_image("new", Subset.TRAINING, 2),
        get_image("new", Subset.TRAINING, 2),
        get_image("new", Subset.TRAINING, 2),
        get_image("new", Subset.TRAINING, 2),
    ]

    new_val = [
        get_image("new", Subset.VALIDATION, 1),
        get_image("new", Subset.VALIDATION, 1),
        get_image("new", Subset.VALIDATION, 1),
        get_image("new", Subset.VALIDATION, 1),
        get_image("new", Subset.VALIDATION, 2),
        get_image("new", Subset.VALIDATION, 2),
        get_image("new", Subset.VALIDATION, 2),
        get_image("new", Subset.VALIDATION, 2),
    ]
    old = old_train + old_val
    new = new_train + new_val
    if data_type == "old":
        return DatasetEntity(old * 3), [labels[0]]
    return DatasetEntity((old * 3 + new * 2)), labels


# pylint: disable=too-many-locals, too-many-statements
def main(args):
    """Main function of Detection Sample."""
    logger.info("[SL] Train initial model with OLD dataset")
    dataset, labels_list = load_test_dataset("old")
    labels_schema = LabelSchemaEntity.from_labels(labels_list)

    logger.info(f"Train dataset: {len(dataset.get_subset(Subset.TRAINING))} items")
    logger.info(f"Validation dataset: {len(dataset.get_subset(Subset.VALIDATION))} items")

    logger.info("Load model template")
    model_template = parse_model_template(args.template_file_path)

    logger.info("Set hyperparameters")
    params = create(model_template.hyper_parameters.data)
    params.learning_parameters.num_iters = 5
    params.learning_parameters.batch_size = 8

    logger.info("Setup environment")
    environment = TaskEnvironment(
        model=None, hyper_parameters=params, label_schema=labels_schema, model_template=model_template
    )

    logger.info("Create base Task")
    task_impl_path = model_template.entrypoints.base
    task_cls = get_task_class(task_impl_path)
    task = task_cls(task_environment=environment)

    logger.info("Train model")
    initial_model = ModelEntity(
        dataset,
        environment.get_model_configuration(),
    )
    task.train(dataset, initial_model)

    logger.info("Class-incremental learning with OLD + NEW dataset")
    dataset, labels_list = load_test_dataset("new")
    labels_schema = LabelSchemaEntity.from_labels(labels_list)

    logger.info(f"Train dataset: {len(dataset.get_subset(Subset.TRAINING))} items")
    logger.info(f"Validation dataset: {len(dataset.get_subset(Subset.VALIDATION))} items")

    logger.info("Load model template")
    model_template = parse_model_template(args.template_file_path)

    logger.info("Set hyperparameters")
    params = create(model_template.hyper_parameters.data)
    params.learning_parameters.num_iters = 5
    params.learning_parameters.batch_size = 8

    logger.info("Setup environment")
    environment = TaskEnvironment(
        model=initial_model, hyper_parameters=params, label_schema=labels_schema, model_template=model_template
    )

    environment.model = ModelEntity(
        train_dataset=dataset,
        configuration=environment.get_model_configuration(),
        model_adapters=initial_model.model_adapters,
    )

    logger.info("Create base Task")
    task_impl_path = model_template.entrypoints.base
    task_cls = get_task_class(task_impl_path)
    task = task_cls(task_environment=environment)

    logger.info("Train model")
    output_model = ModelEntity(
        dataset,
        environment.get_model_configuration(),
    )
    task.train(dataset, output_model)

    logger.info("Get predictions on the validation set")
    validation_dataset = dataset.get_subset(Subset.VALIDATION)
    predicted_validation_dataset = task.infer(
        validation_dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True)
    )
    resultset = ResultSetEntity(
        model=output_model,
        ground_truth_dataset=validation_dataset,
        prediction_dataset=predicted_validation_dataset,
    )
    logger.info("Estimate quality on validation set")
    task.evaluate(resultset)
    logger.info(str(resultset.performance))

    if args.export:
        logger.info("Export model")
        exported_model = ModelEntity(
            dataset,
            environment.get_model_configuration(),
        )
        task.export(ExportType.OPENVINO, exported_model)

        logger.info("Create OpenVINO Task")
        environment.model = exported_model
        openvino_task_impl_path = model_template.entrypoints.openvino
        openvino_task_cls = get_task_class(openvino_task_impl_path)
        openvino_task = openvino_task_cls(environment)

        logger.info("Get predictions on the validation set")
        predicted_validation_dataset = openvino_task.infer(
            validation_dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True)
        )
        resultset = ResultSetEntity(
            model=exported_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )
        logger.info("Estimate quality on validation set")
        openvino_task.evaluate(resultset)
        logger.info(str(resultset.performance))

        # TODO: Need to check POT with openvino 2022.2
        pot_check = "FAILED"
        try:
            # POT
            logger.info("Run POT optimization")
            optimized_model = ModelEntity(
                dataset,
                environment.get_model_configuration(),
            )
            openvino_task.optimize(OptimizationType.POT, dataset, optimized_model, OptimizationParameters())

            logger.info("Get predictions on the validation set")
            predicted_validation_dataset = openvino_task.infer(
                validation_dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True)
            )
            resultset = ResultSetEntity(
                model=optimized_model,
                ground_truth_dataset=validation_dataset,
                prediction_dataset=predicted_validation_dataset,
            )
            logger.info("Performance of optimized model:")
            openvino_task.evaluate(resultset)
            logger.info(str(resultset.performance))
            pot_check = "PASSED"
        except Exception:  # pylint: disable=broad-except
            logger.warning("POT is not working..")

        nncf_check = "FAILED"
        try:
            # NNCF test
            task_impl_path = model_template.entrypoints.nncf
            nncf_task_cls = get_task_class(task_impl_path)

            print("Create NNCF Task")
            environment.model = output_model
            nncf_task_impl_path = model_template.entrypoints.nncf
            nncf_task_cls = get_task_class(nncf_task_impl_path)
            nncf_task = nncf_task_cls(environment)

            optimized_model = ModelEntity(
                dataset,
                configuration=environment.get_model_configuration(),
            )
            nncf_task.optimize(OptimizationType.NNCF, dataset, optimized_model)

            logger.info("Inferring the optimised model on the validation set.")
            predicted_validation_dataset = nncf_task.infer(
                validation_dataset.with_empty_annotations(),
                InferenceParameters(is_evaluation=True),
            )
            resultset = ResultSetEntity(
                model=optimized_model,
                ground_truth_dataset=validation_dataset,
                prediction_dataset=predicted_validation_dataset,
            )

            logger.info("Evaluating the optimized model on the validation set.")
            nncf_task.evaluate(resultset)
            logger.info(str(resultset.performance))

            logger.info("Exporting the model.")
            exported_model = ModelEntity(
                train_dataset=dataset,
                configuration=environment.get_model_configuration(),
            )
            nncf_task.export(ExportType.OPENVINO, exported_model)
            environment.model = exported_model

            logger.info("Creating the OpenVINO Task.")
            environment.model = exported_model
            openvino_task_impl_path = model_template.entrypoints.openvino
            nncf_openvino_task_cls = get_task_class(openvino_task_impl_path)
            nncf_openvino_task = nncf_openvino_task_cls(environment)

            logger.info("Inferring the exported model on the validation set.")
            predicted_validation_dataset = nncf_openvino_task.infer(
                validation_dataset.with_empty_annotations(),
                InferenceParameters(is_evaluation=True),
            )

            logger.info("Evaluating the exported model on the validation set.")
            resultset = ResultSetEntity(
                model=exported_model,
                ground_truth_dataset=validation_dataset,
                prediction_dataset=predicted_validation_dataset,
            )
            nncf_openvino_task.evaluate(resultset)
            logger.info(str(resultset.performance))
            nncf_check = "PASSED"
        except Exception:  # pylint: disable=broad-except
            logger.warning("NNCF is not working")

        logger.info("train: PASSED")
        logger.info("export: PASSED")
        logger.info(f"POT: {pot_check}")
        logger.info(f"NNCF: {nncf_check}")


if __name__ == "__main__":
    sys.exit(main(parse_args()) or 0)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import List

import numpy as np

from ..builder import SCALAR_SCHEDULERS
from .base import BaseScalarScheduler


@SCALAR_SCHEDULERS.register_module()
class StepScalarScheduler(BaseScalarScheduler):
    """Step learning rate scheduler.

    Example:
        >>> scheduler = StepScalarScheduler(scales=[1.0, 0.1, 0.01], num_iters=[100, 200])
        This means that the learning rate will be 1.0 for the first 100 iterations,
        0.1 for the next 200 iterations, and 0.01 for the rest of the iterations.

    Args:
        scales (List[int]): List of learning rate scales.
        num_iters (List[int]): A list specifying the count of iterations at each scale.
        by_epoch (bool): Whether to use epoch as the unit of iteration.
    """

    def __init__(self, scales: List[float], num_iters: List[int], by_epoch: bool = False):
        super(StepScalarScheduler, self).__init__()

        self.by_epoch = by_epoch

        assert len(scales) == len(num_iters) + 1
        assert len(scales) > 0

        self._scales = list(scales)
        self._iter_ranges = list(num_iters) + [np.iinfo(np.int32).max]

    def _get_value(self, step, epoch_size) -> float:
        if step is None:
            return float(self._scales[-1])

        out_scale_idx = 0
        for iter_range in self._iter_ranges:
            if self.by_epoch:
                iter_threshold = epoch_size * iter_range
            else:
                iter_threshold = iter_range

            if step < iter_threshold:
                break

            out_scale_idx += 1

        return float(self._scales[out_scale_idx])



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest

from otx.algorithms.classification.adapters.mmcls.data import (
    OTXClsDataset,
    OTXHierarchicalClsDataset,
    OTXMultilabelClsDataset,
)
from otx.algorithms.classification.adapters.mmcls.data.datasets import SelfSLDataset
from otx.algorithms.classification.utils import get_multihead_class_info
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.classification.test_helper import (
    DEFAULT_CLS_TEMPLATE,
    init_environment,
    setup_configurable_parameters,
)
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


def create_cls_dataset():
    image = Image(data=np.random.randint(low=0, high=255, size=(8, 8, 3)))
    annotation = Annotation(
        shape=Rectangle.generate_full_box(),
        labels=[ScoredLabel(LabelEntity(name="test_selfsl_dataset", domain=Domain.CLASSIFICATION))],
    )
    annotation_scene = AnnotationSceneEntity(annotations=[annotation], kind=AnnotationSceneKind.ANNOTATION)
    dataset_item = DatasetItemEntity(media=image, annotation_scene=annotation_scene)

    dataset = DatasetEntity(items=[dataset_item])
    return dataset, dataset.get_labels()


class TestSelfSLDataset:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.otx_dataset, _ = create_cls_dataset()
        self.pipeline = {
            "view0": [dict(type="ImageToTensor", keys=["img"])],
            "view1": [dict(type="ImageToTensor", keys=["img"])],
        }

    @e2e_pytest_unit
    def test_self_sl_dataset_init_params_validation(self):
        """Test SelfSLDataset initialization parameters validation."""
        correct_values_dict = {
            "otx_dataset": self.otx_dataset,
            "pipeline": self.pipeline,
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "otx_dataset" parameter
            ("otx_dataset", unexpected_str),
            # Unexpected string is specified as "pipeline" parameter
            ("pipeline", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=SelfSLDataset,
        )

    @e2e_pytest_unit
    def test_getitem(self):
        """Test __getitem__ method."""
        dataset = SelfSLDataset(otx_dataset=self.otx_dataset, pipeline=self.pipeline)

        data_item = dataset[0]
        for i in range(1, 3):
            assert f"dataset_item{i}" in data_item
            assert f"width{i}" in data_item
            assert f"height{i}" in data_item
            assert f"index{i}" in data_item
            assert f"filename{i}" in data_item
            assert f"ori_filename{i}" in data_item
            assert f"img{i}" in data_item
            assert f"img_shape{i}" in data_item
            assert f"ori_shape{i}" in data_item
            assert f"pad_shape{i}" in data_item
            assert f"img_norm_cfg{i}" in data_item
            assert f"img_fields{i}" in data_item


class TestOTXClsDataset:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.hyper_parameters, self.model_template = setup_configurable_parameters(DEFAULT_CLS_TEMPLATE)
        self.dataset_len = 20

    @pytest.mark.parametrize(
        "adapter_type",
        [OTXClsDataset, OTXMultilabelClsDataset],
    )
    @e2e_pytest_unit
    def test_create_dataset_adapter(self, adapter_type):
        multilabel = adapter_type == OTXClsDataset
        self.task_environment, self.dataset = init_environment(
            self.hyper_parameters, self.model_template, multilabel, False, self.dataset_len
        )
        dataset = adapter_type(self.dataset, labels=self.dataset.get_labels())
        assert dataset.num_classes == len(self.dataset.get_labels())
        assert len(dataset) == self.dataset_len * dataset.num_classes
        for d in dataset:
            assert d is not None

    @pytest.mark.parametrize(
        "adapter_type",
        [OTXClsDataset, OTXMultilabelClsDataset],
    )
    @e2e_pytest_unit
    def test_metric_dataset_adapter(self, adapter_type):
        multilabel = adapter_type == OTXMultilabelClsDataset
        self.task_environment, self.dataset = init_environment(
            self.hyper_parameters, self.model_template, multilabel, False, self.dataset_len
        )
        dataset = adapter_type(self.dataset, labels=self.dataset.get_labels())
        results = np.ones((len(dataset), dataset.num_classes))
        metrics = dataset.evaluate(results)

        assert len(metrics) > 0
        if multilabel:
            metrics["mAP"] > 0
        else:
            metrics["accuracy"] > 0

    @e2e_pytest_unit
    def test_create_hierarchical_adapter(self):
        self.task_environment, self.dataset = init_environment(
            self.hyper_parameters, self.model_template, False, True, self.dataset_len
        )
        class_info = get_multihead_class_info(self.task_environment.label_schema)
        dataset = OTXHierarchicalClsDataset(
            otx_dataset=self.dataset, labels=self.dataset.get_labels(), hierarchical_info=class_info
        )
        assert dataset.num_classes == len(self.dataset.get_labels())
        assert len(dataset) == self.dataset_len * dataset.num_classes
        for d in dataset:
            assert d is not None

    @e2e_pytest_unit
    def test_metric_hierarchical_adapter(self):
        self.task_environment, self.dataset = init_environment(
            self.hyper_parameters, self.model_template, False, True, self.dataset_len
        )
        class_info = get_multihead_class_info(self.task_environment.label_schema)
        dataset = OTXHierarchicalClsDataset(
            otx_dataset=self.dataset, labels=self.dataset.get_labels(), hierarchical_info=class_info
        )

        results = np.zeros((len(dataset), dataset.num_classes))
        metrics = dataset.evaluate(results)

        assert len(metrics) > 0
        assert metrics["accuracy"] > 0



"""OTX building command 'otx build'.

This command allows you to build an OTX workspace, provide usable backbone configurations,
and build models with new backbone replacements.
"""
# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from pathlib import Path

from otx.cli.manager.config_manager import TASK_TYPE_TO_SUB_DIR_NAME, ConfigManager
from otx.cli.utils.parser import get_parser_and_hprams_data

SUPPORTED_TASKS = (
    "CLASSIFICATION",
    "DETECTION",
    "INSTANCE_SEGMENTATION",
    "ROTATED_DETECTION",
    "SEGMENTATION",
    "ACTION_CLASSIFICATION",
    "ACTION_DETECTION",
    "ANOMALY_CLASSIFICATION",
    "ANOMALY_DETECTION",
    "ANOMALY_SEGMENTATION",
)


def get_args():
    """Parses command line arguments."""
    parser, _, _ = get_parser_and_hprams_data()

    parser.add_argument(
        "--train-data-roots",
        help="Comma-separated paths to training data folders.",
    )
    parser.add_argument(
        "--val-data-roots",
        help="Comma-separated paths to validation data folders.",
    )
    parser.add_argument(
        "--test-data-roots",
        help="Comma-separated paths to test data folders.",
    )
    parser.add_argument(
        "--unlabeled-data-roots",
        help="Comma-separated paths to unlabeled data folders",
    )
    parser.add_argument(
        "--unlabeled-file-list",
        help="Comma-separated paths to unlabeled file list",
    )
    parser.add_argument("--task", help=f"The currently supported options: {SUPPORTED_TASKS}.", default="")
    parser.add_argument(
        "--train-type",
        help=f"The currently supported options: {TASK_TYPE_TO_SUB_DIR_NAME.keys()}.",
        type=str,
        default="incremental",
    )
    parser.add_argument(
        "--work-dir",
        help="Location where the workspace.",
        default=None,
    )
    parser.add_argument(
        "--model", help="Enter the name of the model you want to use. (Ex. EfficientNet-B0).", default=""
    )
    parser.add_argument(
        "--backbone",
        help="Available Backbone Type can be found using 'otx find --backbone {framework}'.\n"
        "If there is an already created backbone configuration yaml file, enter the corresponding path.",
    )

    return parser.parse_args()


def main():
    """Main function for model or backbone or task building."""

    args = get_args()
    config_manager = ConfigManager(args, mode="build")
    if args.task:
        config_manager.task_type = args.task.upper()
    if args.work_dir:
        config_manager.workspace_root = Path(args.work_dir)

    # Auto-Configuration for model template
    config_manager.configure_template(model=args.model)

    config_manager.build_workspace(new_workspace_path=args.work_dir)

    # Auto-Configuration for Dataset configuration
    config_manager.configure_data_config()

    # Build Backbone related
    if args.backbone:
        from otx.cli.builder import Builder

        builder = Builder()
        missing_args = []
        if not args.backbone.endswith((".yml", ".yaml", ".json")):
            backbone_config_file = str(config_manager.workspace_root / "backbone.yaml")
            missing_args = builder.build_backbone_config(args.backbone, backbone_config_file)
        if missing_args:
            print(
                f"[!] {args.backbone} backbone has inputs that the user must enter.\n"
                f"[!] Edit {backbone_config_file} and run 'otx build --backbone backbone.yaml'."
            )
        else:
            builder.merge_backbone(config_manager.workspace_root / "model.py", backbone_config_file)

    return dict(retcode=0, task_type=args.task)


if __name__ == "__main__":
    main()



import os

import pytest

from otx.mpa.seg.inferrer import SegInferrer, replace_ImageToTensor
from otx.mpa.utils.config_utils import MPAConfig
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.segmentation.test_helpers import (
    DEFAULT_RECIPE_CONFIG_PATH,
    DEFAULT_SEG_TEMPLATE_DIR,
)


class TestOTXSegTrainer:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        cfg = MPAConfig.fromfile(DEFAULT_RECIPE_CONFIG_PATH)
        self.inferrer = SegInferrer(name="", mode="train", config=cfg, common_cfg=None, index=0)
        self.model_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "model.py"))
        self.data_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "data_pipeline.py"))

    @e2e_pytest_unit
    def test_run(self, mocker):
        fake_output = {"classes": [1, 2], "eval_predictions": None, "feature_vectors": None}
        mock_infer = mocker.patch.object(SegInferrer, "infer", return_value=fake_output)

        returned_value = self.inferrer.run(self.model_cfg, "", self.data_cfg)
        mock_infer.assert_called_once()
        assert returned_value == {"outputs": fake_output}

    @e2e_pytest_unit
    def test_infer(self, mocker):
        cfg = self.inferrer.configure(self.model_cfg, "", self.data_cfg, training=False)
        mocker.patch.object(SegInferrer, "configure_samples_per_gpu")
        mocker.patch.object(SegInferrer, "configure_compat_cfg")
        mock_infer_callback = mocker.patch.object(SegInferrer, "set_inference_progress_callback")

        returned_value = self.inferrer.infer(cfg)
        mock_infer_callback.assert_called_once()

        assert "classes" in returned_value
        assert "eval_predictions" in returned_value
        assert "feature_vectors" in returned_value
        assert len(returned_value["eval_predictions"]) >= 0


@e2e_pytest_unit
def test_replace_ImageToTensor():
    test_pipeline = [
        dict(type="LoadImageFromFile"),
        dict(
            type="MultiScaleFlipAug",
            transforms=[
                dict(type="Resize", keep_ratio=False),
                dict(type="ImageToTensor", keys=["img"]),
            ],
        ),
        dict(type="ImageToTensor", keys=["img"]),
    ]
    returned_value = replace_ImageToTensor(test_pipeline)

    for pipeline in returned_value:
        if "transforms" in pipeline:
            values = [p["type"] for p in pipeline["transforms"]]
        else:
            values = [pipeline["type"]]
        assert "ImageToTensor" not in values



"""Tests the methods in the OpenVINO task."""

# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from copy import deepcopy

import numpy as np

from otx.algorithms.anomaly.tasks.openvino import OpenVINOTask
from otx.algorithms.anomaly.tasks.train import TrainingTask
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.model import ModelEntity, ModelOptimizationType
from otx.api.entities.model_template import TaskType
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType


class TestOpenVINOTask:
    """Tests methods in the OpenVINO task."""

    def set_normalization_params(self, output_model: ModelEntity):
        """Sets normalization parameters for an untrained output model.

        This is needed as untrained model might have nan values for normalization parameters which will raise an error.
        """
        output_model.set_data("image_threshold", np.float32(0.5).tobytes())
        output_model.set_data("pixel_threshold", np.float32(0.5).tobytes())
        output_model.set_data("min", np.float32(0).tobytes())
        output_model.set_data("max", np.float32(1).tobytes())

    def test_openvino(self, tmpdir, setup_task_environment):
        """Tests the OpenVINO optimize method."""
        root = str(tmpdir.mkdir("anomaly_openvino_test"))

        setup_task_environment = deepcopy(setup_task_environment)  # since fixture is mutable
        task_type = setup_task_environment.task_type
        dataset: DatasetEntity = setup_task_environment.dataset
        task_environment = setup_task_environment.task_environment
        output_model = setup_task_environment.output_model

        # set normalization params for the output model
        train_task = TrainingTask(task_environment, output_path=root)
        self.set_normalization_params(output_model)
        train_task.save_model(output_model)
        task_environment.model = output_model
        train_task.export(ExportType.OPENVINO, output_model)

        # Create OpenVINO task
        openvino_task = OpenVINOTask(task_environment)

        # call inference
        dataset = dataset.get_subset(Subset.VALIDATION)
        predicted_dataset = openvino_task.infer(
            dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True)
        )

        # call evaluate
        result_set = ResultSetEntity(output_model, dataset, predicted_dataset)
        openvino_task.evaluate(result_set)
        if task_type in (TaskType.ANOMALY_CLASSIFICATION, TaskType.ANOMALY_DETECTION):
            assert result_set.performance.score.name == "f-measure"
        elif task_type == TaskType.ANOMALY_SEGMENTATION:
            assert result_set.performance.score.name == "Dice Average"

        # optimize to POT
        openvino_task.optimize(OptimizationType.POT, dataset, output_model, OptimizationParameters())
        assert output_model.optimization_type == ModelOptimizationType.POT
        assert output_model.get_data("label_schema.json") is not None

        # deploy
        openvino_task.deploy(output_model)
        assert output_model.exportable_code is not None
